"""
Functions containing utilities for quality control which
may or may not be program dependent

"""

import os
import numpy as np
import pandas as pd


def calculate_dqr_times(
        obj, variable=None, txt_path=None, threshold=None,
        qc_bit=None, return_missing=True):
    """
    Function to retrieve start and end times of missing or bad data. Function
    will retrieve start and end time strings in a format that the DQR
    submission tool can read, print them to the console, and save to a .txt
    file if desired.

    Parameters
    ----------
    obj : ACT Object
        Xarray Dataset as read by ACT where data are stored.
    variable : str or list of str
        Data variable(s) in the object to check. Can check multiple variables.
    txt_path : str
        Full path to directory in which to save .txt files with start and end
        times. If directory doesn't exist the program will create it. If set
        to None then no .txt files will be created. Default is None.
    threshold : int
        Threshold of number of data points to trigger start and end time
        calculations. Value is interpreted differently depending on the
        resolution of the data in the specified files. For example, if data is
        1-minute data, a threshold of 30 would mean flagged data more than 30
        minutes apart would show up as two separate time ranges; if data is 30-
        minute data, a threshold of 2 would mean flagged data more than 60 mins
        apart would show up as two separate time ranges.
    qc_bit : int
        Bit number to choose if finding times for bad data. If set then will
        override searching for missing data. Default is None.
    return_missing : bool
        Specifies whether or not to return times of data flagged as missing. If
        set to False, will return times of bad data. Default is True.

    Returns
    -------
    time_strings : list
        List of tuples with the first element as the start time and the second
        element as the end time.

    """
    # Determine if searching for either bad or missing data
    if not return_missing:
        return_bad = True
        return_missing = False
    else:
        return_bad = False

    # If qc bit set then make sure searching for bad data
    if qc_bit:
        return_bad = True
        return_missing = False

    # Clean files. Converts from ARM to CF standards
    obj.clean.cleanup(cleanup_arm_qc=True,
                      handle_missing_value=True,
                      link_qc_variables=True)
    date = obj.attrs['_file_dates'][0]
    datastream = obj.attrs['_datastream']

    # Make variable instance a list always
    if variable is not None:
        if not isinstance(variable, list):
            variable = [variable]
    else:
        variable = []

    # Make sure that threshold number is an integer. If not convert to
    # closest integer
    if threshold is not None:
        if not isinstance(threshold, int):
            int(round(threshold))
    else:
        print('You must specify a threshold for separating ranges of' +
              ' flagged data')
        return

    # If return_missing then search for indices where data is equal to
    # missing value.
    if return_missing:
        for var in variable:
            # Get indices where data is being flagged as missing
            idx = np.where(np.isnan(obj[var].values))[0]

            # Find where there are gaps in flagged data
            time_diff = np.diff(idx)

            # Find indices in flagged data where gaps occur
            splits = np.where(time_diff > threshold)
            splits = splits[0] + 1

            # If no bad indices then exit
            if len(idx) == 0:
                print('No missing data for {} on '.format(var) + date)
                continue
            else:
                idx = np.split(idx, splits)
            # Now that we have all of the stretches of bad flags, get
            # corresponding datetimes from time data
            dt_times = []

            for time in idx:
                # If there is only one flagged data point skip
                if len(time) < 2:
                    pass
                else:
                    dt_times.append((obj['time'].values[time[0]],
                                    obj['time'].values[time[-1]]))
            # Convert the datetimes to strings
            time_strings = []
            for st, et in dt_times:
                start_time = pd.to_datetime(str(st)).strftime(
                    '%Y-%m-%d %H:%M:%S')
                end_time = pd.to_datetime(str(et)).strftime(
                    '%Y-%m-%d %H:%M:%S')
                time_strings.append((start_time, end_time))
                # Print times to screen
                print('Missing Data for {} begins at: '.format(var) +
                      start_time)
                print('Missing Data for {} ends at: '.format(var) + end_time)
            if txt_path:
                _write_dqr_times_to_txt(datastream, date, txt_path, var,
                                        time_strings)
            return time_strings

    # If return_bad then search for times in the corresponding qc variable
    # where the flags are tripped
    if return_bad:
        for var in variable:
            # If a1 level data, return.
            if 'a1' in obj.attrs['data_level']:
                print('No QC is present in a1 level.')
                return
            # Get QC data from corresponding QC variable
            qc_var = 'qc_' + var
            try:
                qc_data = obj[qc_var].values
            except KeyError:
                print('Unable to calculate start and end times for bad data')
                continue
            # Make sure qc bit is an integer
            if not isinstance(qc_bit, int):
                raise TypeError('QC bit must be an integer')
            # Get indices where data is being flagged for given qc bit
            idx = np.where(qc_data == 2 ** (qc_bit - 1))[0]

            # Find where there are gaps in flagged data
            time_diff = np.diff(idx)

            # Find indices in flagged data where gaps occur
            splits = np.where(time_diff > threshold)
            splits = splits[0] + 1

            # If no bad indices then exit
            if len(idx) == 0:
                print('No bad data on ' + date + ' for selected QC bit for' +
                      ' variable ' + var)
                continue
            else:
                idx = np.split(idx, splits)
            # Now that we have all of the stretches of bad flags, get
            # corresponding datetimes from time data
            dt_times = []

            for time in idx:
                # If there is only one flagged data point skip
                if len(time) < 2:
                    pass
                else:
                    dt_times.append((obj['time'].values[time[0]],
                                    obj['time'].values[time[-1]]))
            # Convert the datetimes to strings
            time_strings = []
            for st, et in dt_times:
                start_time = pd.to_datetime(str(st)).strftime(
                    '%Y-%m-%d %H:%M:%S')
                end_time = pd.to_datetime(str(et)).strftime(
                    '%Y-%m-%d %H:%M:%S')
                time_strings.append((start_time, end_time))
                # Print times to screen
                print('Bad Data for {} Begins at: '.format(var) + start_time)
                print('Bad Data for {} Ends at: '.format(var) + end_time)
            if txt_path:
                _write_dqr_times_to_txt(datastream, date, txt_path, var,
                                        time_strings)
            return time_strings


def _write_dqr_times_to_txt(datastream, date, txt_path, variable,
                            time_strings):
    """
    Writes flagged data time range(s) to a .txt file. The naming convention is
    dqrtimes_datastream.date.txt.

    Parameters
    ----------
    datastream : str
        ARM datastream name (ie, sgpmetE13.b1).
    date : str
        Date of time range(s) being written.
    txt_path : str
        Base path of where the .txt files are being written.
    variable : str
        Name of variable being flagged.
    time_strings : list of tuples
        List of every start and end time to be written.

    """
    print('Writing data to text file for ' + datastream + ' ' + variable +
          ' on ' + date + ' at ' + txt_path + '...', flush=True)
    full_path = txt_path + '/' + datastream
    if os.path.exists(full_path) is False:
        os.mkdir(full_path)
    with open(full_path + '/dqrtimes_' + datastream + '.' + date + '.' +
              variable + '.txt', 'w') as text_file:
        for st, et in time_strings:
            text_file.write('%s, ' % st)
            text_file.write('%s \n' % et)



"""
Stores the class for SkewTDisplay.

"""

# Import third party libraries
import matplotlib.pyplot as plt
import numpy as np
import warnings
import scipy

try:
    from pkg_resources import DistributionNotFound
    import metpy
    import metpy.calc as mpcalc
    METPY_AVAILABLE = True
except ImportError:
    METPY_AVAILABLE = False
except (ModuleNotFoundError, DistributionNotFound):
    warnings.warn("MetPy is installed but could not be imported. " +
                  "Please check your MetPy installation. Some features " +
                  "will be disabled.", ImportWarning)
    METPY_AVAILABLE = False

# Import Local Libs
from ..utils import datetime_utils as dt_utils
from copy import deepcopy
from .plot import Display

if METPY_AVAILABLE:
    from metpy.units import units
    from metpy.plots import SkewT


class SkewTDisplay(Display):
    """
    A class for making Skew-T plots.

    This is inherited from the :func:`act.plotting.Display`
    class and has therefore has the same attributes as that class.
    See :func:`act.plotting.Display`
    for more information. There are no additional attributes or parameters
    to this class.

    In order to create Skew-T plots, ACT needs the MetPy package to be
    installed on your system. More information about
    MetPy go here: https://unidata.github.io/MetPy/latest/index.html.

    Examples
    --------
    Here is an example of how to make a Skew-T plot using ACT:

    .. code-block :: python

        sonde_ds = act.io.armfiles.read_netcdf(
           act.tests.sample_files.EXAMPLE_SONDE1)

        skewt = act.plotting.SkewTDisplay(sonde_ds)
        skewt.plot_from_u_and_v('u_wind', 'v_wind', 'pres', 'tdry', 'dp')
        plt.show()

    """
    def __init__(self, obj, subplot_shape=(1,), ds_name=None, **kwargs):
        # We want to use our routine to handle subplot adding, not the main
        # one
        if not METPY_AVAILABLE:
            raise ImportError("MetPy need to be installed on your system to " +
                              "make Skew-T plots.")
        new_kwargs = kwargs.copy()
        super().__init__(obj, None, ds_name,
                         subplot_kw=dict(projection='skewx'), **new_kwargs)

        # Make a SkewT object for each subplot
        self.add_subplots(subplot_shape, **kwargs)

    def add_subplots(self, subplot_shape=(1,), **kwargs):
        """
        Adds subplots to the Display object. The current
        figure in the object will be deleted and overwritten.

        Parameters
        ----------
        subplot_shape : 1 or 2D tuple, list, or array
            The structure of the subplots in (rows, cols).
        subplot_kw : dict, optional
            The kwargs to pass into fig.subplots.
        **kwargs : keyword arguments
            Any other keyword arguments that will be passed
            into :func:`matplotlib.pyplot.figure` when the figure
            is made. The figure is only made if the *fig*
            property is None. See the matplotlib
            documentation for further details on what keyword
            arguments are available.

        """
        del self.axes
        if self.fig is None:
            self.fig = plt.figure(**kwargs)
        self.SkewT = np.empty(shape=subplot_shape, dtype=SkewT)
        self.axes = np.empty(shape=subplot_shape, dtype=plt.Axes)
        if len(subplot_shape) == 1:
            for i in range(subplot_shape[0]):
                subplot_tuple = (subplot_shape[0], 1, i + 1)
                self.SkewT[i] = SkewT(fig=self.fig, subplot=subplot_tuple)
                self.axes[i] = self.SkewT[i].ax
        elif len(subplot_shape) == 2:
            for i in range(subplot_shape[0]):
                for j in range(subplot_shape[1]):
                    subplot_tuple = (subplot_shape[0],
                                     subplot_shape[1],
                                     i * subplot_shape[1] + j + 1)
                    self.SkewT[i, j] = SkewT(fig=self.fig, subplot=subplot_tuple)
                    self.axes[i, j] = self.SkewT[i, j].ax
        else:
            raise ValueError("Subplot shape must be 1 or 2D!")

    def set_xrng(self, xrng, subplot_index=(0,)):
        """
        Sets the x range of the plot.

        Parameters
        ----------
        xrng : 2 number array.
            The x limits of the plot.
        subplot_index : 1 or 2D tuple, list, or array
            The index of the subplot to set the x range of.

        """
        if self.axes is None:
            raise RuntimeError("set_xrng requires the plot to be displayed.")

        if not hasattr(self, 'xrng') or np.all(self.xrng == 0):
            if len(self.axes.shape) == 2:
                self.xrng = np.zeros((self.axes.shape[0], self.axes.shape[1], 2))
            else:
                self.xrng = np.zeros((self.axes.shape[0], 2))

        self.axes[subplot_index].set_xlim(xrng)
        self.xrng[subplot_index, :] = np.array(xrng)

    def set_yrng(self, yrng, subplot_index=(0,)):
        """
        Sets the y range of the plot.

        Parameters
        ----------
        yrng : 2 number array
            The y limits of the plot.
        subplot_index : 1 or 2D tuple, list, or array
            The index of the subplot to set the x range of.

        """
        if self.axes is None:
            raise RuntimeError("set_yrng requires the plot to be displayed.")

        if not hasattr(self, 'yrng') or np.all(self.yrng == 0):
            if len(self.axes.shape) == 2:
                self.yrng = np.zeros((self.axes.shape[0], self.axes.shape[1], 2))
            else:
                self.yrng = np.zeros((self.axes.shape[0], 2))

        if not hasattr(self, 'yrng') and len(self.axes.shape) == 2:
            self.yrng = np.zeros((self.axes.shape[0], self.axes.shape[1], 2))
        elif not hasattr(self, 'yrng') and len(self.axes.shape) == 1:
            self.yrng = np.zeros((self.axes.shape[0], 2))

        if yrng[0] == yrng[1]:
            yrng[1] = yrng[1] + 1

        self.axes[subplot_index].set_ylim(yrng)
        self.yrng[subplot_index, :] = yrng

    def plot_from_spd_and_dir(self, spd_field, dir_field,
                              p_field, t_field, td_field, dsname=None,
                              **kwargs):
        """
        This plot will make a sounding plot from wind data that is given
        in speed and direction.

        Parameters
        ----------
        spd_field : str
            The name of the field corresponding to the wind speed.
        dir_field : str
            The name of the field corresponding to the wind direction
            in degrees from North.
        p_field : str
            The name of the field containing the atmospheric pressure.
        t_field : str
            The name of the field containing the atmospheric temperature.
        td_field : str
            The name of the field containing the dewpoint.
        dsname : str or None
            The name of the datastream to plot. Set to None to make ACT
            attempt to automatically determine this.
        kwargs : dict
            Additional keyword arguments will be passed into
            :func:`act.plotting.SkewTDisplay.plot_from_u_and_v`

        Returns
        -------
        ax : matplotlib axis handle
            The matplotlib axis handle corresponding to the plot.

        """
        if dsname is None and len(self._obj.keys()) > 1:
            raise ValueError(("You must choose a datastream when there are 2 "
                              "or more datasets in the TimeSeriesDisplay "
                              "object."))
        elif dsname is None:
            dsname = list(self._obj.keys())[0]

        # Make temporary field called tempu, tempv
        spd = self._obj[dsname][spd_field].values
        dir = self._obj[dsname][dir_field].values
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=RuntimeWarning)
            tempu = -np.sin(np.deg2rad(dir)) * spd
            tempv = -np.cos(np.deg2rad(dir)) * spd
        self._obj[dsname]["temp_u"] = deepcopy(self._obj[dsname][spd_field])
        self._obj[dsname]["temp_v"] = deepcopy(self._obj[dsname][spd_field])
        self._obj[dsname]["temp_u"].values = tempu
        self._obj[dsname]["temp_v"].values = tempv
        the_ax = self.plot_from_u_and_v("temp_u", "temp_v", p_field,
                                        t_field, td_field, dsname, **kwargs)
        del self._obj[dsname]["temp_u"], self._obj[dsname]["temp_v"]
        return the_ax

    def plot_from_u_and_v(self, u_field, v_field, p_field,
                          t_field, td_field, dsname=None, subplot_index=(0,),
                          p_levels_to_plot=None, show_parcel=True,
                          shade_cape=True, shade_cin=True, set_title=None,
                          plot_barbs_kwargs=dict(), plot_kwargs=dict(),):
        """
        This function will plot a Skew-T from a sounding dataset. The wind
        data must be given in u and v.

        Parameters
        ----------
        u_field : str
            The name of the field containing the u component of the wind.
        v_field : str
            The name of the field containing the v component of the wind.
        p_field : str
            The name of the field containing the pressure.
        t_field : str
            The name of the field containing the temperature.
        td_field : str
            The name of the field containing the dewpoint temperature.
        dsname : str or None
            The name of the datastream to plot. Set to None to make ACT
            attempt to automatically determine this.
        subplot_index : tuple
            The index of the subplot to make the plot on.
        p_levels_to_plot : 1D array
            The pressure levels to plot the wind barbs on. Set to None
            to have ACT to use neatly spaced defaults of
            50, 100, 200, 300, 400, 500, 600, 700, 750, 800,
            850, 900, 950, and 1000 hPa.
        show_parcel : bool
            Set to True to show the temperature of a parcel lifted
            from the surface.
        shade_cape : bool
            Set to True to shade the CAPE red.
        shade_cin : bool
            Set to True to shade the CIN blue.
        set_title : None or str
            The title of the plot is set to this. Set to None to use
            a default title.
        plot_barbs_kwargs : dict
            Additional keyword arguments to pass into MetPy's
            SkewT.plot_barbs.
        plot_kwargs : dict
            Additional keyword arguments to pass into MetPy's
            SkewT.plot.

        Returns
        -------
        ax : matplotlib axis handle
            The axis handle to the plot.

        """
        if dsname is None and len(self._obj.keys()) > 1:
            raise ValueError(("You must choose a datastream when there are 2 "
                              "or more datasets in the TimeSeriesDisplay "
                              "object."))
        elif dsname is None:
            dsname = list(self._obj.keys())[0]

        if p_levels_to_plot is None:
            p_levels_to_plot = np.array([50., 100., 200., 300., 400.,
                                         500., 600., 700., 750., 800.,
                                         850., 900., 950., 1000.])
        T = self._obj[dsname][t_field]
        T_units = self._obj[dsname][t_field].attrs["units"]
        if T_units == "C":
            T_units = "degC"

        T = T.values * getattr(units, T_units)
        Td = self._obj[dsname][td_field]
        Td_units = self._obj[dsname][td_field].attrs["units"]
        if Td_units == "C":
            Td_units = "degC"

        Td = Td.values * getattr(units, Td_units)
        u = self._obj[dsname][u_field]
        u_units = self._obj[dsname][u_field].attrs["units"]
        u = u.values * getattr(units, u_units)

        v = self._obj[dsname][v_field]
        v_units = self._obj[dsname][v_field].attrs["units"]
        v = v.values * getattr(units, v_units)

        p = self._obj[dsname][p_field]
        p_units = self._obj[dsname][p_field].attrs["units"]
        p = p.values * getattr(units, p_units)

        u_red = np.zeros_like(p_levels_to_plot) * getattr(units, u_units)
        v_red = np.zeros_like(p_levels_to_plot) * getattr(units, v_units)
        p_levels_to_plot = p_levels_to_plot * getattr(units, p_units)
        for i in range(len(p_levels_to_plot)):
            index = np.argmin(np.abs(p_levels_to_plot[i] - p))
            u_red[i] = u[index].magnitude * getattr(units, u_units)
            v_red[i] = v[index].magnitude * getattr(units, v_units)

        u_red = u_red.magnitude
        v_red = v_red.magnitude
        self.SkewT[subplot_index].plot(p, T, 'r', **plot_kwargs)
        self.SkewT[subplot_index].plot(p, Td, 'g', **plot_kwargs)
        self.SkewT[subplot_index].plot_barbs(
            p_levels_to_plot, u_red, v_red, **plot_barbs_kwargs)

        # Metpy fix if Pressure does not decrease monotonically in
        # your sounding.
        try:
            prof = mpcalc.parcel_profile(p, T[0], Td[0]).to('degC')
        except metpy.calc.exceptions.InvalidSoundingError:
            p = scipy.ndimage.median_filter(p, 3, output=float)
            p = metpy.units.units.Quantity(p, p_units)
            prof = mpcalc.parcel_profile(p, T[0], Td[0]).to('degC')

        if show_parcel:
            # Only plot where prof > T
            lcl_pressure, lcl_temperature = mpcalc.lcl(p[0], T[0], Td[0])
            self.SkewT[subplot_index].plot(
                lcl_pressure, lcl_temperature, 'ko', markerfacecolor='black',
                **plot_kwargs)
            self.SkewT[subplot_index].plot(
                p, prof, 'k', linewidth=2, **plot_kwargs)

        if shade_cape:
            self.SkewT[subplot_index].shade_cape(
                p, T, prof, linewidth=2)

        if shade_cin:
            self.SkewT[subplot_index].shade_cin(
                p, T, prof, linewidth=2)

        # Set Title
        if set_title is None:
            set_title = ' '.join(
                [dsname, 'on',
                 dt_utils.numpy_to_arm_date(self._obj[dsname].time.values[0])])

        self.axes[subplot_index].set_title(set_title)

        # Set Y Limit
        our_data = p.magnitude
        if np.isfinite(our_data).any():
            yrng = [np.nanmax(our_data), np.nanmin(our_data)]
        else:
            yrng = [1000., 100.]
        self.set_yrng(yrng, subplot_index)

        # Set X Limit
        xrng = [np.nanmin(T.magnitude) - 10., np.nanmax(T.magnitude) + 10.]
        self.set_xrng(xrng, subplot_index)

        return self.axes[subplot_index]



""" Unit tests for ACT utils module. """

from datetime import datetime
import pytz

import act
import numpy as np
import pandas as pd
import pytest
import xarray as xr
import tempfile
from pathlib import Path


def test_dates_between():
    start_date = '20190101'
    end_date = '20190110'
    date_list = act.utils.dates_between(start_date, end_date)
    answer = [datetime(2019, 1, 1),
              datetime(2019, 1, 2),
              datetime(2019, 1, 3),
              datetime(2019, 1, 4),
              datetime(2019, 1, 5),
              datetime(2019, 1, 6),
              datetime(2019, 1, 7),
              datetime(2019, 1, 8),
              datetime(2019, 1, 9),
              datetime(2019, 1, 10)]
    assert date_list == answer


def test_add_in_nan():
    # Make a 1D array with a 4 day gap in the data
    time_list = [np.datetime64(datetime(2019, 1, 1, 1, 0)),
                 np.datetime64(datetime(2019, 1, 1, 1, 1)),
                 np.datetime64(datetime(2019, 1, 1, 1, 2)),
                 np.datetime64(datetime(2019, 1, 1, 1, 8)),
                 np.datetime64(datetime(2019, 1, 1, 1, 9))]
    data = np.linspace(0., 8., 5)

    time_list = xr.DataArray(time_list)
    data = xr.DataArray(data)
    time_filled, data_filled = act.utils.add_in_nan(
        time_list, data)

    assert(data_filled.data[8] == 6.)

    time_answer = [np.datetime64(datetime(2019, 1, 1, 1, 0)),
                   np.datetime64(datetime(2019, 1, 1, 1, 1)),
                   np.datetime64(datetime(2019, 1, 1, 1, 2)),
                   np.datetime64(datetime(2019, 1, 1, 1, 3)),
                   np.datetime64(datetime(2019, 1, 1, 1, 4)),
                   np.datetime64(datetime(2019, 1, 1, 1, 5)),
                   np.datetime64(datetime(2019, 1, 1, 1, 6)),
                   np.datetime64(datetime(2019, 1, 1, 1, 7)),
                   np.datetime64(datetime(2019, 1, 1, 1, 8)),
                   np.datetime64(datetime(2019, 1, 1, 1, 9))]

    assert(time_filled[8].values == time_answer[8])
    assert(time_filled[5].values == time_answer[5])

    time_filled, data_filled = act.utils.add_in_nan(
        time_list[0], data[0])
    assert time_filled.values == time_list[0]
    assert data_filled.values == data[0]


def test_get_missing_value():
    obj = act.io.armfiles.read_netcdf(act.tests.sample_files.EXAMPLE_EBBR1)
    missing = act.utils.data_utils.get_missing_value(
        obj, 'latent_heat_flux', use_FillValue=True, add_if_missing_in_obj=True)
    assert missing == -9999

    obj['latent_heat_flux'].attrs['missing_value'] = -9998
    missing = act.utils.data_utils.get_missing_value(obj, 'latent_heat_flux')
    assert missing == -9998


def test_convert_units():
    obj = act.io.armfiles.read_netcdf(act.tests.sample_files.EXAMPLE_EBBR1)
    data = obj['soil_temp_1'].values
    in_units = obj['soil_temp_1'].attrs['units']
    r_data = act.utils.data_utils.convert_units(data, in_units, 'K')
    assert np.ceil(r_data[0]) == 285

    data = act.utils.data_utils.convert_units(r_data, 'K', 'C')
    assert np.ceil(data[0]) == 12

    try:
        obj.utils.change_units()
    except ValueError as error:
        assert str(error) == "Need to provide 'desired_unit' keyword for .change_units() method"

    desired_unit = 'degF'
    skip_vars = [ii for ii in obj.data_vars if ii.startswith('qc_')]
    obj.utils.change_units(variables=None, desired_unit=desired_unit,
                           skip_variables=skip_vars, skip_standard=True)
    units = []
    for var_name in obj.data_vars:
        try:
            units.append(obj[var_name].attrs['units'])
        except KeyError:
            pass
    indices = [i for i, x in enumerate(units) if x == desired_unit]
    assert indices == [0, 2, 4, 6, 8, 32, 34, 36, 38, 40]

    var_name = 'home_signal_15'
    desired_unit = 'V'
    obj.utils.change_units(var_name, desired_unit, skip_variables='lat')
    assert obj[var_name].attrs['units'] == desired_unit

    var_names = ['home_signal_15', 'home_signal_30']
    obj.utils.change_units(var_names, desired_unit)
    for var_name in var_names:
        assert obj[var_name].attrs['units'] == desired_unit

    obj.close()
    del obj

    obj = act.io.armfiles.read_netcdf(act.tests.sample_files.EXAMPLE_CEIL1)
    var_name = 'range'
    desired_unit = 'km'
    obj = obj.utils.change_units(var_name, desired_unit)
    assert obj[var_name].attrs['units'] == desired_unit
    assert np.isclose(np.sum(obj[var_name].values), 952.56, atol=0.01)

    obj.close()
    del obj


def test_ts_weighted_average():
    obj = act.io.armfiles.read_netcdf(act.tests.sample_files.EXAMPLE_MET_WILDCARD)
    cf_ds = {'sgpmetE13.b1': {'variable': ['tbrg_precip_total', 'org_precip_rate_mean',
                                           'pwd_precip_rate_mean_1min'],
                              'weight': [0.8, 0.15, 0.05], 'object': obj}}
    data = act.utils.data_utils.ts_weighted_average(cf_ds)

    np.testing.assert_almost_equal(np.sum(data), 84.9, decimal=1)


def test_accum_precip():
    obj = act.io.armfiles.read_netcdf(
        act.tests.sample_files.EXAMPLE_MET_WILDCARD)

    obj = act.utils.accumulate_precip(obj, 'tbrg_precip_total')
    dmax = round(np.nanmax(obj['tbrg_precip_total_accumulated']))
    assert dmax == 13.0

    obj = act.utils.accumulate_precip(obj, 'tbrg_precip_total', time_delta=60)
    dmax = round(np.nanmax(obj['tbrg_precip_total_accumulated']))
    assert dmax == 13.0

    obj['tbrg_precip_total'].attrs['units'] = 'mm/hr'
    obj = act.utils.accumulate_precip(obj, 'tbrg_precip_total')
    dmax = np.round(np.nanmax(obj['tbrg_precip_total_accumulated']), 2)
    assert dmax == 0.22


def test_calc_cog_sog():
    obj = act.io.armfiles.read_netcdf(
        act.tests.sample_files.EXAMPLE_NAV)

    obj = act.utils.calc_cog_sog(obj)

    cog = obj['course_over_ground'].values
    sog = obj['speed_over_ground'].values

    np.testing.assert_almost_equal(cog[10], 170.987, decimal=3)
    np.testing.assert_almost_equal(sog[15], 0.448, decimal=3)

    obj = obj.rename({'lat': 'latitude', 'lon': 'longitude'})
    obj = act.utils.calc_cog_sog(obj)
    np.testing.assert_almost_equal(cog[10], 170.987, decimal=3)
    np.testing.assert_almost_equal(sog[15], 0.448, decimal=3)


def test_destination_azimuth_distance():
    lat = 37.1509
    lon = -98.362
    lat2, lon2 = act.utils.destination_azimuth_distance(lat, lon, 180., 100)

    np.testing.assert_almost_equal(lat2, 37.150, decimal=3)
    np.testing.assert_almost_equal(lon2, -98.361, decimal=3)


def test_calculate_dqr_times():
    ebbr1_ds = act.io.armfiles.read_netcdf(
        act.tests.sample_files.EXAMPLE_EBBR1)
    ebbr2_ds = act.io.armfiles.read_netcdf(
        act.tests.sample_files.EXAMPLE_EBBR2)
    brs_ds = act.io.armfiles.read_netcdf(
        act.tests.sample_files.EXAMPLE_BRS)
    ebbr1_result = act.utils.calculate_dqr_times(
        ebbr1_ds, variable=['soil_temp_1'], threshold=2)
    ebbr2_result = act.utils.calculate_dqr_times(
        ebbr2_ds, variable=['rh_bottom_fraction'],
        qc_bit=3, threshold=2)
    ebbr3_result = act.utils.calculate_dqr_times(
        ebbr2_ds, variable=['rh_bottom_fraction'],
        qc_bit=3)
    brs_result = act.utils.calculate_dqr_times(
        brs_ds, variable='down_short_hemisp_min', qc_bit=2, threshold=30)
    assert ebbr1_result == [('2019-11-25 02:00:00', '2019-11-25 04:30:00')]
    assert ebbr2_result == [('2019-11-30 00:00:00', '2019-11-30 11:00:00')]
    assert brs_result == [('2019-07-05 01:57:00', '2019-07-05 11:07:00')]
    assert ebbr3_result is None
    with tempfile.TemporaryDirectory() as tmpdirname:
        write_file = Path(tmpdirname)
        brs_result = act.utils.calculate_dqr_times(brs_ds, variable='down_short_hemisp_min',
                                                   qc_bit=2, threshold=30, txt_path=str(write_file))

    brs_result = act.utils.calculate_dqr_times(brs_ds, variable='down_short_hemisp_min',
                                               qc_bit=2, threshold=30, return_missing=False)
    assert len(brs_result[0]) == 2

    ebbr1_ds.close()
    ebbr2_ds.close()
    brs_ds.close()


def test_decode_present_weather():
    obj = act.io.armfiles.read_netcdf(act.tests.sample_files.EXAMPLE_MET1)
    obj = act.utils.decode_present_weather(obj, variable='pwd_pw_code_inst')

    data = obj['pwd_pw_code_inst_decoded'].values
    result = 'No significant weather observed'
    assert data[0] == result
    assert data[100] == result
    assert data[600] == result

    np.testing.assert_raises(ValueError, act.utils.inst_utils.decode_present_weather, obj)
    np.testing.assert_raises(ValueError, act.utils.inst_utils.decode_present_weather,
                             obj, variable='temp_temp')


def test_datetime64_to_datetime():
    time_datetime = [datetime(2019, 1, 1, 1, 0),
                     datetime(2019, 1, 1, 1, 1),
                     datetime(2019, 1, 1, 1, 2),
                     datetime(2019, 1, 1, 1, 3),
                     datetime(2019, 1, 1, 1, 4)]

    time_datetime64 = [np.datetime64(datetime(2019, 1, 1, 1, 0)),
                       np.datetime64(datetime(2019, 1, 1, 1, 1)),
                       np.datetime64(datetime(2019, 1, 1, 1, 2)),
                       np.datetime64(datetime(2019, 1, 1, 1, 3)),
                       np.datetime64(datetime(2019, 1, 1, 1, 4))]

    time_datetime64_to_datetime = act.utils.datetime_utils.datetime64_to_datetime(time_datetime64)
    assert time_datetime == time_datetime64_to_datetime


def test_create_pyart_obj():
    try:
        obj = act.io.mpl.read_sigma_mplv5(act.tests.EXAMPLE_SIGMA_MPLV5)
    except Exception:
        return

    radar = act.utils.create_pyart_obj(obj, range_var='range')
    variables = list(radar.fields)
    assert 'nrb_copol' in variables
    assert 'nrb_crosspol' in variables
    assert radar.sweep_start_ray_index['data'][-1] == 67
    assert radar.sweep_end_ray_index['data'][-1] == 101
    assert radar.fixed_angle['data'] == 2.0
    assert radar.scan_type == 'ppi'
    assert radar.sweep_mode['data'] == 'ppi'
    np.testing.assert_allclose(
        radar.sweep_number['data'][-3:],
        [1., 1., 1.])
    np.testing.assert_allclose(
        radar.sweep_number['data'][0:3],
        [0., 0., 0.])

    # coordinates
    np.testing.assert_allclose(
        radar.azimuth['data'][0:5],
        [-95., -92.5, -90., -87.5, -85.])
    np.testing.assert_allclose(
        radar.elevation['data'][0:5],
        [2., 2., 2., 2., 2.])
    np.testing.assert_allclose(
        radar.range['data'][0:5],
        [14.98962308, 44.96886923, 74.94811538,
         104.92736153, 134.90660768])
    gate_lat = radar.gate_latitude['data'][0, 0:5]
    gate_lon = radar.gate_longitude['data'][0, 0:5]
    gate_alt = radar.gate_altitude['data'][0, 0:5]
    np.testing.assert_allclose(
        gate_lat, [38.95293483, 38.95291135, 38.95288786,
                   38.95286437, 38.95284089])
    np.testing.assert_allclose(
        gate_lon, [-76.8363515, -76.83669666, -76.83704182,
                   -76.83738699, -76.83773215])
    np.testing.assert_allclose(
        gate_alt, [62.84009906, 63.8864653, 64.93293721,
                   65.9795148, 67.02619806])
    obj.close()
    del radar


def test_add_solar_variable():
    obj = act.io.armfiles.read_netcdf(act.tests.EXAMPLE_NAV)
    new_obj = act.utils.geo_utils.add_solar_variable(obj)

    assert 'sun_variable' in list(new_obj.keys())
    assert new_obj['sun_variable'].values[10] == 1
    assert np.sum(new_obj['sun_variable'].values) >= 598

    new_obj = act.utils.geo_utils.add_solar_variable(obj, dawn_dusk=True)
    assert 'sun_variable' in list(new_obj.keys())
    assert new_obj['sun_variable'].values[10] == 1
    assert np.sum(new_obj['sun_variable'].values) >= 1234

    obj = act.io.armfiles.read_netcdf(act.tests.EXAMPLE_MET1)
    new_obj = act.utils.geo_utils.add_solar_variable(obj, dawn_dusk=True)
    assert np.sum(new_obj['sun_variable'].values) >= 1046

    obj = act.io.armfiles.read_netcdf(act.tests.EXAMPLE_IRTSST)
    obj = obj.fillna(0)
    new_obj = act.utils.geo_utils.add_solar_variable(obj)
    assert np.sum(new_obj['sun_variable'].values) >= 12

    obj = act.io.armfiles.read_netcdf(act.tests.EXAMPLE_IRTSST)
    obj.drop_vars('lat')
    pytest.raises(
        ValueError, act.utils.geo_utils.add_solar_variable, obj)

    obj = act.io.armfiles.read_netcdf(act.tests.EXAMPLE_IRTSST)
    obj.drop_vars('lon')
    pytest.raises(
        ValueError, act.utils.geo_utils.add_solar_variable, obj)
    obj.close()
    new_obj.close()


def test_reduce_time_ranges():
    time = pd.date_range(start='2020-01-01T00:00:00', freq='1min', periods=100)
    time = time.to_list()
    time = time[0:50] + time[60:]
    result = act.utils.datetime_utils.reduce_time_ranges(time)
    assert len(result) == 2
    assert result[1][1].minute == 39

    result = act.utils.datetime_utils.reduce_time_ranges(time, broken_barh=True)
    assert len(result) == 2


def test_planck_converter():
    wnum = 1100
    temp = 300
    radiance = 81.5
    result = act.utils.radiance_utils.planck_converter(wnum=wnum, temperature=temp)
    np.testing.assert_almost_equal(result, radiance, decimal=1)
    result = act.utils.radiance_utils.planck_converter(wnum=wnum, radiance=radiance)
    assert np.ceil(result) == temp
    np.testing.assert_raises(ValueError, act.utils.radiance_utils.planck_converter)


def test_solar_azimuth_elevation():

    obj = act.io.armfiles.read_netcdf(act.tests.EXAMPLE_NAV)

    elevation, azimuth, distance = act.utils.geo_utils.get_solar_azimuth_elevation(
        latitude=obj['lat'].values[0], longitude=obj['lon'].values[0], time=obj['time'].values,
        library='skyfield', temperature_C='standard', pressure_mbar='standard')
    assert np.isclose(np.nanmean(elevation), 26.408, atol=0.001)
    assert np.isclose(np.nanmean(azimuth), 179.732, atol=0.001)
    assert np.isclose(np.nanmean(distance), 0.985, atol=0.001)


def test_get_sunrise_sunset_noon():

    obj = act.io.armfiles.read_netcdf(act.tests.EXAMPLE_NAV)

    sunrise, sunset, noon = act.utils.geo_utils.get_sunrise_sunset_noon(
        latitude=obj['lat'].values[0], longitude=obj['lon'].values[0],
        date=obj['time'].values[0], library='skyfield')
    assert sunrise[0].replace(microsecond=0) == datetime(2018, 1, 31, 22, 36, 32)
    assert sunset[0].replace(microsecond=0) == datetime(2018, 2, 1, 17, 24, 4)
    assert noon[0].replace(microsecond=0) == datetime(2018, 2, 1, 8, 2, 10)

    sunrise, sunset, noon = act.utils.geo_utils.get_sunrise_sunset_noon(
        latitude=obj['lat'].values[0], longitude=obj['lon'].values[0],
        date=obj['time'].values[0], library='skyfield', timezone=True)
    assert sunrise[0].replace(microsecond=0) == datetime(2018, 1, 31, 22, 36, 32, tzinfo=pytz.UTC)
    assert sunset[0].replace(microsecond=0) == datetime(2018, 2, 1, 17, 24, 4, tzinfo=pytz.UTC)
    assert noon[0].replace(microsecond=0) == datetime(2018, 2, 1, 8, 2, 10, tzinfo=pytz.UTC)

    sunrise, sunset, noon = act.utils.geo_utils.get_sunrise_sunset_noon(
        latitude=obj['lat'].values[0], longitude=obj['lon'].values[0],
        date='20180201', library='skyfield')
    assert sunrise[0].replace(microsecond=0) == datetime(2018, 1, 31, 22, 36, 32)
    assert sunset[0].replace(microsecond=0) == datetime(2018, 2, 1, 17, 24, 4)
    assert noon[0].replace(microsecond=0) == datetime(2018, 2, 1, 8, 2, 10)

    sunrise, sunset, noon = act.utils.geo_utils.get_sunrise_sunset_noon(
        latitude=obj['lat'].values[0], longitude=obj['lon'].values[0],
        date=['20180201'], library='skyfield')
    assert sunrise[0].replace(microsecond=0) == datetime(2018, 1, 31, 22, 36, 32)
    assert sunset[0].replace(microsecond=0) == datetime(2018, 2, 1, 17, 24, 4)
    assert noon[0].replace(microsecond=0) == datetime(2018, 2, 1, 8, 2, 10)

    sunrise, sunset, noon = act.utils.geo_utils.get_sunrise_sunset_noon(
        latitude=obj['lat'].values[0], longitude=obj['lon'].values[0],
        date=datetime(2018, 2, 1), library='skyfield')
    assert sunrise[0].replace(microsecond=0) == datetime(2018, 1, 31, 22, 36, 32)
    assert sunset[0].replace(microsecond=0) == datetime(2018, 2, 1, 17, 24, 4)
    assert noon[0].replace(microsecond=0) == datetime(2018, 2, 1, 8, 2, 10)

    sunrise, sunset, noon = act.utils.geo_utils.get_sunrise_sunset_noon(
        latitude=obj['lat'].values[0], longitude=obj['lon'].values[0],
        date=datetime(2018, 2, 1, tzinfo=pytz.UTC), library='skyfield')
    assert sunrise[0].replace(microsecond=0) == datetime(2018, 1, 31, 22, 36, 32)
    assert sunset[0].replace(microsecond=0) == datetime(2018, 2, 1, 17, 24, 4)
    assert noon[0].replace(microsecond=0) == datetime(2018, 2, 1, 8, 2, 10)

    sunrise, sunset, noon = act.utils.geo_utils.get_sunrise_sunset_noon(
        latitude=obj['lat'].values[0], longitude=obj['lon'].values[0],
        date=[datetime(2018, 2, 1)], library='skyfield')
    assert sunrise[0].replace(microsecond=0) == datetime(2018, 1, 31, 22, 36, 32)
    assert sunset[0].replace(microsecond=0) == datetime(2018, 2, 1, 17, 24, 4)
    assert noon[0].replace(microsecond=0) == datetime(2018, 2, 1, 8, 2, 10)

    sunrise, sunset, noon = act.utils.geo_utils.get_sunrise_sunset_noon(
        latitude=85.0, longitude=-140., date=[datetime(2018, 6, 1)], library='skyfield')
    assert sunrise[0].replace(microsecond=0) == datetime(2018, 3, 30, 10, 48, 48)
    assert sunset[0].replace(microsecond=0) == datetime(2018, 9, 12, 8, 50, 14)
    assert noon[0].replace(microsecond=0) == datetime(2018, 6, 1, 21, 17, 52)


def test_is_sun_visible():
    obj = act.io.armfiles.read_netcdf(act.tests.sample_files.EXAMPLE_EBBR1)
    result = act.utils.geo_utils.is_sun_visible(
        latitude=obj['lat'].values, longitude=obj['lon'].values,
        date_time=obj['time'].values)
    assert len(result) == 48
    assert sum(result) == 20

    result = act.utils.geo_utils.is_sun_visible(
        latitude=obj['lat'].values, longitude=obj['lon'].values,
        date_time=obj['time'].values[0])
    assert result == [False]

    result = act.utils.geo_utils.is_sun_visible(
        latitude=obj['lat'].values, longitude=obj['lon'].values,
        date_time=[datetime(2019, 11, 25, 13, 30, 00)])
    assert result == [True]

    result = act.utils.geo_utils.is_sun_visible(
        latitude=obj['lat'].values, longitude=obj['lon'].values,
        date_time=datetime(2019, 11, 25, 13, 30, 00))
    assert result == [True]


def test_convert_to_potential_temp():
    obj = act.io.armfiles.read_netcdf(act.tests.sample_files.EXAMPLE_MET1)

    temp_var_name = 'temp_mean'
    press_var_name = 'atmos_pressure'
    temp = act.utils.data_utils.convert_to_potential_temp(
        obj, temp_var_name, press_var_name=press_var_name)
    assert np.isclose(np.nansum(temp), -4240.092, rtol=0.001, atol=0.001)
    temp = act.utils.data_utils.convert_to_potential_temp(
        temperature=obj[temp_var_name].values, pressure=obj[press_var_name].values,
        temp_var_units=obj[temp_var_name].attrs['units'],
        press_var_units=obj[press_var_name].attrs['units'])
    assert np.isclose(np.nansum(temp), -4240.092, rtol=0.001, atol=0.0011)

    with np.testing.assert_raises(ValueError):
        temp = act.utils.data_utils.convert_to_potential_temp(
            temperature=obj[temp_var_name].values, pressure=obj[press_var_name].values,
            temp_var_units=obj[temp_var_name].attrs['units'])

    with np.testing.assert_raises(ValueError):
        temp = act.utils.data_utils.convert_to_potential_temp(
            temperature=obj[temp_var_name].values, pressure=obj[press_var_name].values,
            press_var_units=obj[press_var_name].attrs['units'])


def test_height_adjusted_temperature():
    obj = act.io.armfiles.read_netcdf(act.tests.sample_files.EXAMPLE_MET1)

    temp_var_name = 'temp_mean'
    press_var_name = 'atmos_pressure'
    temp = act.utils.data_utils.height_adjusted_temperature(
        obj, temp_var_name, height_difference=100, height_units='m',
        press_var_name=press_var_name)
    assert np.isclose(np.nansum(temp), -6834.291, rtol=0.001, atol=0.001)

    temp = act.utils.data_utils.height_adjusted_temperature(
        obj, temp_var_name=temp_var_name, height_difference=-900,
        height_units='feet')
    assert np.isclose(np.nansum(temp), -1904.7257, rtol=0.001, atol=0.001)

    temp = act.utils.data_utils.height_adjusted_temperature(
        obj, temp_var_name, height_difference=-200,
        height_units='m', press_var_name=press_var_name,
        pressure=102.325, press_var_units='kPa')
    assert np.isclose(np.nansum(temp), -2871.5435, rtol=0.001, atol=0.001)

    temp = act.utils.data_utils.height_adjusted_temperature(
        height_difference=25.2, height_units='m',
        temperature=obj[temp_var_name].values,
        temp_var_units=obj[temp_var_name].attrs['units'],
        pressure=obj[press_var_name].values,
        press_var_units=obj[press_var_name].attrs['units'])
    assert np.isclose(np.nansum(temp), -5847.511, rtol=0.001, atol=0.001)

    with np.testing.assert_raises(ValueError):
        temp = act.utils.data_utils.height_adjusted_temperature(
            height_difference=25.2, height_units='m',
            temperature=obj[temp_var_name].values,
            temp_var_units=None,
            pressure=obj[press_var_name].values,
            press_var_units=obj[press_var_name].attrs['units'])


def test_height_adjusted_pressure():
    obj = act.io.armfiles.read_netcdf(act.tests.sample_files.EXAMPLE_MET1)

    press_var_name = 'atmos_pressure'
    temp = act.utils.data_utils.height_adjusted_pressure(
        obj=obj, press_var_name=press_var_name, height_difference=20,
        height_units='m')
    assert np.isclose(np.nansum(temp), 142020.83, rtol=0.001, atol=0.001)

    temp = act.utils.data_utils.height_adjusted_pressure(
        height_difference=-100, height_units='ft',
        pressure=obj[press_var_name].values,
        press_var_units=obj[press_var_name].attrs['units'])
    assert np.isclose(np.nansum(temp), 142877.69, rtol=0.001, atol=0.001)

    with np.testing.assert_raises(ValueError):
        temp = act.utils.data_utils.height_adjusted_pressure(
            height_difference=-100, height_units='ft',
            pressure=obj[press_var_name].values,
            press_var_units=None)



"""
This module contains classes for displaying data.
:func:`act.plotting.Display` is the
base class on which all other Display classes are inherited from. If you are making
a new Display object, please make it inherited from this class.

| :func:`act.plotting.ContourDisplay` handles the plotting of contour plots.
| :func:`act.plotting.GeographicPlotDisplay` handles the plotting of lat-lon plots.
| :func:`act.plotting.HistogramDisplay` handles the plotting of histogram plots.
| :func:`act.plotting.SkewTDisplay` handles the plotting of Skew-T diagrams.
| :func:`act.plotting.TimeSeriesDisplay` handles the plotting of timeseries.
| :func:`act.plotting.WindRoseDisplay` handles the plotting of wind rose plots.
| :func:`act.plotting.XSectionDisplay` handles the plotting of cross sections.

"""

from .plot import Display
from .timeseriesdisplay import TimeSeriesDisplay
from .contourdisplay import ContourDisplay
from .windrosedisplay import WindRoseDisplay
from .skewtdisplay import SkewTDisplay
from .xsectiondisplay import XSectionDisplay
from .geodisplay import GeographicPlotDisplay
from .histogramdisplay import HistogramDisplay
from . import common
from . import act_cmap



"""
This module contains sample files used for testing the ARM Community Toolkit.
Files in this module should only be used for testing, not production.

"""

from .sample_files import (EXAMPLE_SONDE1, EXAMPLE_LCL1, EXAMPLE_MET_CSV,
                           EXAMPLE_SONDE_WILDCARD, EXAMPLE_MET1,
                           EXAMPLE_METE40, EXAMPLE_TWP_SONDE_20060121,
                           EXAMPLE_MET_WILDCARD, EXAMPLE_CEIL1,
                           EXAMPLE_CEIL_WILDCARD, EXAMPLE_ANL_CSV,
                           EXAMPLE_MPL_1SAMPLE, EXAMPLE_IRT25m20s,
                           EXAMPLE_MET_CONTOUR, EXAMPLE_NAV,
                           EXAMPLE_AOSMET, EXAMPLE_DLPPI, EXAMPLE_EBBR1,
                           EXAMPLE_EBBR2, EXAMPLE_BRS, EXAMPLE_AERI,
                           EXAMPLE_MFRSR, EXAMPLE_SURFSPECALB1MLAWER,
                           EXAMPLE_SIGMA_MPLV5, EXAMPLE_RL1,
                           EXAMPLE_CO2FLX4M, EXAMPLE_SIRS, EXAMPLE_IRTSST,
                           EXAMPLE_MET_TEST1, EXAMPLE_MET_TEST2,
                           EXAMPLE_STAMP_WILDCARD, EXAMPLE_NOAA_PSL)



"""
Functions for radiosonde related calculations.

"""

import warnings
import numpy as np
import pandas as pd
import xarray as xr
from act.utils.data_utils import convert_to_potential_temp

try:
    from pkg_resources import DistributionNotFound
    import metpy.calc as mpcalc
    METPY_AVAILABLE = True
except ImportError:
    METPY_AVAILABLE = False
except (ModuleNotFoundError, DistributionNotFound):
    warnings.warn("MetPy is installed but could not be imported. " +
                  "Please check your MetPy installation. Some features " +
                  "will be disabled.", ImportWarning)
    METPY_AVAILABLE = False

if METPY_AVAILABLE:
    from metpy.units import units


def calculate_precipitable_water(ds, temp_name='tdry', rh_name='rh',
                                 pres_name='pres'):
    """

    Function to calculate precipitable water vapor from ARM sondewnpn b1 data.
    Will first calculate saturation vapor pressure of all data using Arden-Buck
    equations, then calculate specific humidity and integrate over all pressure
    levels to give us a precipitable water value in centimeters.

    ds : ACT object
        Object as read in by the ACT netCDF reader.
    temp_name : str
        Name of temperature field to use. Defaults to 'tdry' for sondewnpn b1
        level data.
    rh_name : str
        Name of relative humidity field to use. Defaults to 'rh' for sondewnpn
        b1 level data.
    pres_name : str
        Name of atmospheric pressure field to use. Defaults to 'pres' for
        sondewnpn b1 level data.

    """
    temp = ds[temp_name].values
    rh = ds[rh_name].values
    pres = ds[pres_name].values

    # Get list of temperature values for saturation vapor pressure calc
    temperature = []
    for t in np.nditer(temp):
        temperature.append(t)

    # Apply Arden-Buck equation to get saturation vapor pressure
    sat_vap_pres = []
    for t in temperature:
        # Over liquid water, above freezing
        if t >= 0:
            sat_vap_pres.append(0.61121 * np.exp((18.678 - (t / 234.5)) *
                                (t / (257.14 + t))))
        # Over ice, below freezing
        else:
            sat_vap_pres.append(0.61115 * np.exp((23.036 - (t / 333.7)) *
                                (t / (279.82 + t))))

    # convert rh from % to decimal
    rel_hum = []
    for r in np.nditer(rh):
        rel_hum.append(r / 100.)

    # get vapor pressure from rh and saturation vapor pressure
    vap_pres = []
    for i in range(0, len(sat_vap_pres)):
        es = rel_hum[i] * sat_vap_pres[i]
        vap_pres.append(es)

    # Get list of pressure values for mixing ratio calc
    pressure = []
    for p in np.nditer(pres):
        pressure.append(p)

    # Mixing ratio calc

    mix_rat = []
    for i in range(0, len(vap_pres)):
        mix_rat.append(0.622 * vap_pres[i] / (pressure[i] - vap_pres[i]))

    # Specific humidity

    spec_hum = []
    for rat in mix_rat:
        spec_hum.append(rat / (1 + rat))

    # Integrate specific humidity

    pwv = 0.0
    for i in range(1, len(pressure) - 1):
        pwv = pwv + 0.5 * (spec_hum[i] + spec_hum[i - 1]) * (pressure[i - 1] -
                                                             pressure[i])

    pwv = pwv / 0.098
    return pwv


def calculate_stability_indicies(ds, temp_name="temperature",
                                 td_name="dewpoint_temperature",
                                 p_name="pressure",
                                 rh_name='relative_humidity',
                                 moving_ave_window=0):
    """
    Function for calculating stability indices from sounding data.

    Parameters
    ----------
    ds : ACT dataset
        The dataset to compute the stability indicies of. Must have
        temperature, dewpoint, and pressure in vertical coordinates.
    temp_name : str
        The name of the temperature field.
    td_name : str
        The name of the dewpoint field.
    p_name : str
        The name of the pressure field.
    rh_name : str
        The name of the relative humidity field.
    moving_ave_window : int
        Number of points to do a moving average on sounding data to reduce
        noise. This is useful if noise in the sounding is preventing parcel
        ascent.

    Returns
    -------
    ds : ACT dataset
        An ACT dataset with additional stability indicies added.

    """
    if not METPY_AVAILABLE:
        raise ImportError("MetPy need to be installed on your system to " +
                          "calculate stability indices")

    t = ds[temp_name]
    td = ds[td_name]
    p = ds[p_name]
    rh = ds[rh_name]

    if not hasattr(t, "units"):
        raise AttributeError("Temperature field must have units" +
                             " for ACT to discern!")

    if not hasattr(td, "units"):
        raise AttributeError("Dewpoint field must have units" +
                             " for ACT to discern!")

    if not hasattr(p, "units"):
        raise AttributeError("Pressure field must have units" +
                             " for ACT to discern!")
    if t.units == "C":
        t_units = units.degC
    else:
        t_units = getattr(units, t.units)

    if td.units == "C":
        td_units = units.degC
    else:
        td_units = getattr(units, td.units)

    p_units = getattr(units, p.units)
    rh_units = getattr(units, rh.units)

    # Sort all values by decreasing pressure
    t_sorted = np.array(t.values)
    td_sorted = np.array(td.values)
    p_sorted = np.array(p.values)
    rh_sorted = np.array(rh.values)
    ind_sort = np.argsort(p_sorted)
    t_sorted = t_sorted[ind_sort[-1:0:-1]]
    td_sorted = td_sorted[ind_sort[-1:0:-1]]
    p_sorted = p_sorted[ind_sort[-1:0:-1]]
    rh_sorted = rh_sorted[ind_sort[-1:0:-1]]

    if moving_ave_window > 0:
        t_sorted = np.convolve(
            t_sorted, np.ones((moving_ave_window,)) / moving_ave_window)
        td_sorted = np.convolve(
            td_sorted, np.ones((moving_ave_window,)) / moving_ave_window)
        p_sorted = np.convolve(
            p_sorted, np.ones((moving_ave_window,)) / moving_ave_window)
        rh_sorted = np.convolve(
            rh_sorted, np.ones((moving_ave_window,)) / moving_ave_window)

    t_sorted = t_sorted * t_units
    td_sorted = td_sorted * td_units
    p_sorted = p_sorted * p_units
    rh_sorted = rh_sorted * rh_units

    # Calculate mixing ratio
    mr = mpcalc.mixing_ratio_from_relative_humidity(
        p_sorted, t_sorted, rh_sorted)

    # Discussion of issue #361 use virtual temperature.
    vt = mpcalc.virtual_temperature(t_sorted, mr)

    t_profile = mpcalc.parcel_profile(
        p_sorted, t_sorted[0], td_sorted[0])

    # Calculate parcel trajectory
    ds["parcel_temperature"] = t_profile.magnitude
    ds["parcel_temperature"].attrs['units'] = t_profile.units

    # Calculate CAPE, CIN, LCL
    sbcape, sbcin = mpcalc.surface_based_cape_cin(
        p_sorted, vt, td_sorted)

    lcl = mpcalc.lcl(
        p_sorted[0], t_sorted[0], td_sorted[0])
    try:
        lfc = mpcalc.lfc(
            p_sorted[0], t_sorted[0], td_sorted[0])
    except IndexError:
        lfc = np.nan * p_sorted.units

    mucape, mucin = mpcalc.most_unstable_cape_cin(
        p_sorted, vt, td_sorted)

    where_500 = np.argmin(np.abs(p_sorted - 500 * units.hPa))
    li = t_sorted[where_500] - t_profile[where_500]

    ds["surface_based_cape"] = sbcape.magnitude
    ds["surface_based_cape"].attrs['units'] = "J/kg"
    ds["surface_based_cape"].attrs['long_name'] = "Surface-based CAPE"
    ds["surface_based_cin"] = sbcin.magnitude
    ds["surface_based_cin"].attrs['units'] = "J/kg"
    ds["surface_based_cin"].attrs['long_name'] = "Surface-based CIN"
    ds["most_unstable_cape"] = mucape.magnitude
    ds["most_unstable_cape"].attrs['units'] = "J/kg"
    ds["most_unstable_cape"].attrs['long_name'] = "Most unstable CAPE"
    ds["most_unstable_cin"] = mucin.magnitude
    ds["most_unstable_cin"].attrs['units'] = "J/kg"
    ds["most_unstable_cin"].attrs['long_name'] = "Most unstable CIN"
    ds["lifted_index"] = li.magnitude
    ds["lifted_index"].attrs['units'] = t_profile.units
    ds["lifted_index"].attrs['long_name'] = "Lifted index"
    ds["level_of_free_convection"] = lfc.magnitude
    ds["level_of_free_convection"].attrs['units'] = lfc.units
    ds["level_of_free_convection"].attrs['long_name'] = "Level of free convection"
    ds["lifted_condensation_level_temperature"] = lcl[1].magnitude
    ds["lifted_condensation_level_temperature"].attrs['units'] = lcl[1].units
    ds["lifted_condensation_level_temperature"].attrs[
        'long_name'] = "Lifted condensation level temperature"
    ds["lifted_condensation_level_pressure"] = lcl[0].magnitude
    ds["lifted_condensation_level_pressure"].attrs['units'] = lcl[0].units
    ds["lifted_condensation_level_pressure"].attrs[
        'long_name'] = "Lifted condensation level pressure"
    return ds


def calculate_pbl_liu_liang(ds, temperature='tdry', pressure='pres', windspeed='wspd', height='alt',
                            smooth_height=3, land_parameter=True, llj_max_alt=1500., llj_max_wspd=2.):
    """
    Function for calculating the PBL height from a radiosonde profile
    using the Liu-Liang 2010 technique.  There are some slight descrepencies
    in the function from the ARM implementation 1.) it imposes a 1500m (keyword)
    height on the definition of the LLJ and 2.) the interpolation is slightly different
    using python functions

    Parameters
    ----------
    ds : xarray Dataset
        Dataset housing radiosonde profile for calculations
    temperature : str
        The name of the temperature field.
    pressure : str
        The name of the pressure field.
    windspeed : str
        The name of the  wind speed field.
    height : str
        The name of the height field
    smooth_height : int
        Number of points to do a moving average on sounding height data to reduce noise
    land_parameter : boolean
        Set to True if retrievals over land or false to retrievals over water
    llj_max_alt : float
        Maximum altitude the LLJ 2 m/s difference should be checked against
    llj_max_wspd : float
        Maximum wind speed threshold to use to define LLJ

    Returns
    -------
    obj : xarray Dataset
        xarray dataset with results stored in pblht_liu_liang variable

    References
    ----------
    Liu, Shuyan, and Xin-Zhong Liang. "Observed diurnal cycle climatology of planetary
        boundary layer height." Journal of Climate 23, no. 21 (2010): 5790-5809.

    Sivaraman, C., S. McFarlane, E. Chapman, M. Jensen, T. Toto, S. Liu, and M. Fischer.
        "Planetary boundary layer (PBL) height value added product (VAP): Radiosonde retrievals."
        Department of Energy Office of Science Atmospheric Radiation Measurement (ARM) Program
        (United States) (2013).

    """

    time_0 = ds['time'].values
    temp_0 = ds[temperature].values

    ds[pressure] = ds[pressure].rolling(time=smooth_height, min_periods=1, center=True).mean()
    obj = ds.swap_dims(dims_dict={'time': pressure})
    for var in obj:
        obj[var].attrs = ds[var].attrs

    base = 5  # 5 mb base
    starting_pres = base * np.ceil(float(obj[pressure].values[2]) / base)
    p_grid = np.flip(np.arange(100., starting_pres + base, base))

    try:
        obj = obj.sel(pres=p_grid, method='nearest')
    except Exception:
        ds[pressure] = ds[pressure].rolling(time=smooth_height + 4, min_periods=2, center=True).mean()
        obj = ds.swap_dims(dims_dict={'time': pressure})
        for var in obj:
            obj[var].attrs = ds[var].attrs
        try:
            obj = obj.sel(pres=p_grid, method='nearest')
        except Exception:
            raise ValueError('Sonde profile does not have unique pressures after smoothing')

    # Get Data Variables
    if smooth_height > 0:
        alt = pd.Series(obj[height].values).rolling(window=smooth_height, min_periods=0).mean().values
    else:
        alt = obj[height].values
    if np.isnan(alt[0]):
        idx = np.where(~np.isnan(alt))[0]
        agl = alt - alt[idx[0]]
    else:
        agl = alt - alt[0]
    pres = obj[pressure].values
    temp = obj[temperature].values
    wspd = obj[windspeed].values

    # Perform Pre-processing checks
    if len(temp) == 0:
        raise ValueError("No data in profile")

    if np.nanmax(alt) < 1000.:
        raise ValueError("Max altitude < 1000m")

    if np.nanmax(pres) <= 200.:
        raise ValueError("Max pressure <= 200 hPa")

    # Check temperature delta
    t1 = time_0[0]
    t2 = t1 + np.timedelta64(10, 's')
    idx = np.where((time_0 >= t1) & (time_0 <= t2))[0]
    t_delta = abs(temp_0[idx[-1]] - temp_0[idx[0]])
    if t_delta > 30.:
        raise ValueError('Temperature changes by >30º in first 10 seconds')

    # Check min/max
    if np.nanmax(temp) > 50. or np.nanmin(temp) < -90:
        raise ValueError('Temperature outside acceptable range (-90, 50)')

    if np.isnan(pres[0]) or np.isnan(pres[1]):
        raise ValueError('First two pressure values bad')

    # Calculate potential temperature and subsequent gradients
    theta = convert_to_potential_temp(obj=obj, temp_var_name=temperature, press_var_name='pres') + 273.15
    atts = {'units': 'K', 'long_name': 'Potential temperature'}
    da = xr.DataArray(theta, coords=obj['tdry'].coords, dims=obj[temperature].dims, attrs=atts)
    obj['potential_temperature'] = da

    theta_diff = theta[4] - theta[1]
    theta_gradient = np.diff(theta) / np.diff(alt / 1000.)

    # Set up threshold values
    if land_parameter:
        stability_thresh = 1.0  # K
        inst_thresh = 0.5  # K
        overshoot_thresh = 4.0  # K/km
    else:
        stability_thresh = 0.2  # K
        inst_thresh = 0.1  # K
        overshoot_thresh = 0.5  # K/km

    # Check Regimes
    if theta_diff < 0 - stability_thresh:
        regime = 'CBL'
    if theta_diff > abs(stability_thresh):
        regime = 'SBL'
    if (0 - stability_thresh) <= theta_diff <= abs(stability_thresh):
        regime = 'NRL'

    # Calculate for CBL/NRL regimes
    pbl_stable = np.nan
    pbl_shear = np.nan

    if regime == 'CBL' or regime == 'NRL':
        # Calculate gradient from first level
        theta_gradient_0 = theta - theta[0]

        # Only process data above 150m ARM
        idx = np.where(agl > 150)[0][0]
        theta_gradient_0[0:idx] = np.nan

        # Scan upward to find lowest level that meets condition
        idx = np.where(theta_gradient_0 >= inst_thresh)[0]
        theta_gradient[0:idx[0]] = np.nan

        # Scan upward from previous level to search for overlying inversion layer
        idx = np.where(theta_gradient >= overshoot_thresh)[0]
        pbl = alt[idx[0]]
    else:
        idx = np.array([i for i, t in enumerate(theta_gradient[1:-1]) if theta_gradient[i] <
                        theta_gradient[i - 1] and theta_gradient[i] < theta_gradient[i + 1]])

        for i in idx:
            cond1 = (theta_gradient[i] - theta_gradient[i - 1]) < -40.
            cond2 = (theta_gradient[i + 1] < overshoot_thresh) or (theta_gradient[i + 2] < overshoot_thresh)
            if cond1 or cond2:
                # This gets the ARM answer
                pbl_stable = (alt[i + 1] + alt[i]) / 2.
                # pbl_stable = alt[i]
                break

        # Check for low-level jet
        # Find the height of the maximum windspeed and look up to find layer 2m/s lower
        # Stull 1988 indicates LLJ is defined as where there is a relative wind speed
        # maximum that is more than 2 m/s faster than the wind speeds above it within
        # the lowest 1500m of the atmosphere. Keywords to adjust are provided
        idh = np.where(alt <= llj_max_alt)[0]
        max_wspd_ind = [i for i, w in enumerate(wspd[:-1]) if wspd[i] > wspd[i + 1]][0]
        diff = wspd[max_wspd_ind] - wspd[max_wspd_ind:idh[-1]]
        idx = np.where(diff > llj_max_wspd)[0]
        if len(idx) > 0:
            wspd_to_surf = np.diff(np.flip(wspd[0:max_wspd_ind]))
            wspd_monotonic = np.all(wspd_to_surf <= 0.)
            if wspd_monotonic:
                pbl_shear = alt[max_wspd_ind]

        if ~np.all(np.isnan([pbl_stable, pbl_shear])):
            pbl = np.nanmin([pbl_stable, pbl_shear])
        else:
            pbl = -9999.

    atts = {'units': 'm', 'long_name': 'Planteary Boundary Layer Height Liu-Liang'}
    da = xr.DataArray(pbl, attrs=atts)
    ds['pblht_liu_liang'] = da

    atts = {'units': '', 'long_name': 'Planteary Boundary Layer Regime Classification Liu-Liang'}
    da = xr.DataArray(regime, attrs=atts)
    ds['pblht_regime_liu_liang'] = da

    atts = {'units': 'mb', 'long_name': 'Gridded pressure'}
    da = xr.DataArray(pres, coords={'atm_pres_ss': pres}, dims=['atm_pres_ss'], attrs=atts)
    ds['atm_pres_ss'] = da

    atts = {'units': 'K', 'long_name': 'Gridded potential temperature'}
    da = xr.DataArray(theta, coords={'atm_pres_ss': pres}, dims=['atm_pres_ss'], attrs=atts)
    ds['potential_temperature_ss'] = da

    atts = {'units': 'm', 'long_name': 'Gridded altitude'}
    da = xr.DataArray(alt, coords={'atm_pres_ss': pres}, dims=['atm_pres_ss'], attrs=atts)
    ds['alt_ss'] = da

    atts = {'units': 'm', 'long_name': 'PBL Stable Condition 1'}
    da = xr.DataArray(pbl_stable, attrs=atts)
    ds['pblht_liu_liang_stable_cond'] = da

    atts = {'units': 'm', 'long_name': 'PBL Shear Condition 2'}
    da = xr.DataArray(pbl_shear, attrs=atts)
    ds['pblht_liu_liang_shear_cond'] = da

    return ds



"""
Sample data file for use in testing. These files should only
be used for testing ACT.

"""

import os

DATA_PATH = os.path.join(os.path.dirname(__file__), 'data')

EXAMPLE_MET1 = os.path.join(DATA_PATH, 'sgpmetE13.b1.20190101.000000.cdf')
EXAMPLE_MET_CSV = os.path.join(DATA_PATH, 'sgpmetE13.*csv')
EXAMPLE_METE40 = os.path.join(DATA_PATH, 'sgpmetE40.b1.20190508.000000.cdf')
EXAMPLE_CEIL1 = os.path.join(DATA_PATH, 'sgpceilC1.b1.20190101.000000.nc')
EXAMPLE_SONDE1 = os.path.join(DATA_PATH,
                              'sgpsondewnpnC1.b1.20190101.053200.cdf')
EXAMPLE_LCL1 = os.path.join(DATA_PATH, 'met_lcl.nc')
EXAMPLE_SONDE_WILDCARD = os.path.join(DATA_PATH, 'sgpsondewnpn*.cdf')
EXAMPLE_MET_WILDCARD = os.path.join(DATA_PATH, 'sgpmet*201901*.cdf')
EXAMPLE_MET_CONTOUR = os.path.join(DATA_PATH, 'sgpmet*20190508*.cdf')
EXAMPLE_CEIL_WILDCARD = os.path.join(DATA_PATH, 'sgpceil*.cdf')
EXAMPLE_TWP_SONDE_WILDCARD = os.path.join(DATA_PATH, 'twpsondewnpn*.cdf')
EXAMPLE_TWP_SONDE_20060121 = os.path.join(DATA_PATH, 'twpsondewnpn*20060121*.cdf')
EXAMPLE_ANL_CSV = os.path.join(DATA_PATH, 'anltwr_mar19met.data')
EXAMPLE_VISST = os.path.join(
    DATA_PATH, 'twpvisstgridirtemp.c1.20050705.002500.nc')
EXAMPLE_MPL_1SAMPLE = os.path.join(DATA_PATH,
                                   'sgpmplpolfsC1.b1.20190502.000000.cdf')
EXAMPLE_IRT25m20s = os.path.join(DATA_PATH,
                                 'sgpirt25m20sC1.a0.20190601.000000.cdf')
EXAMPLE_NAV = os.path.join(DATA_PATH,
                           'marnavM1.a1.20180201.000000.nc')
EXAMPLE_AOSMET = os.path.join(DATA_PATH,
                              'maraosmetM1.a1.20180201.000000.nc')
EXAMPLE_DLPPI = os.path.join(DATA_PATH, 'sgpdlppiC1.b1.20191015.120023.cdf')
EXAMPLE_DLPPI_MULTI = os.path.join(DATA_PATH, 'sgpdlppiC1.b1.20191015.*.cdf')
EXAMPLE_EBBR1 = os.path.join(DATA_PATH, 'sgp30ebbrE32.b1.20191125.000000.nc')
EXAMPLE_EBBR2 = os.path.join(DATA_PATH, 'sgp30ebbrE32.b1.20191130.000000.nc')
EXAMPLE_BRS = os.path.join(DATA_PATH, 'sgpbrsC1.b1.20190705.000000.cdf')
EXAMPLE_AERI = os.path.join(DATA_PATH, 'sgpaerich1C1.b1.20190501.000342.nc')
EXAMPLE_IRTSST = os.path.join(DATA_PATH, 'marirtsstM1.b1.20190320.000000.nc')
EXAMPLE_MFRSR = os.path.join(DATA_PATH, 'sgpmfrsr7nchE38.b1.20190514.180000.nc')
EXAMPLE_SURFSPECALB1MLAWER = os.path.join(
    DATA_PATH, 'nsasurfspecalb1mlawerC1.c1.20160609.080000.nc')
EXAMPLE_SIGMA_MPLV5 = os.path.join(DATA_PATH, '201509021500.bi')
EXAMPLE_RL1 = os.path.join(DATA_PATH, 'sgprlC1.a0.20160131.000000.nc')
EXAMPLE_CO2FLX4M = os.path.join(DATA_PATH, 'sgpco2flx4mC1.b1.20201007.001500.nc')
EXAMPLE_SIRS = os.path.join(DATA_PATH, 'sgpsirsE13.b1.20190101.000000.cdf')
EXAMPLE_GML_RADIATION = os.path.join(DATA_PATH, 'brw21001.dat')
EXAMPLE_GML_MET = os.path.join(DATA_PATH, 'met_brw_insitu_1_obop_hour_2020.txt')
EXAMPLE_GML_OZONE = os.path.join(DATA_PATH, 'brw_12_2020_hour.dat')
EXAMPLE_GML_CO2 = os.path.join(DATA_PATH, 'co2_brw_surface-insitu_1_ccgg_MonthlyData.txt')
EXAMPLE_GML_HALO = os.path.join(DATA_PATH, 'brw_CCl4_Day.dat')
EXAMPLE_MET_TEST1 = os.path.join(DATA_PATH, 'sgpmet_no_time.nc')
EXAMPLE_MET_TEST2 = os.path.join(DATA_PATH, 'sgpmet_test_time.nc')
EXAMPLE_STAMP_WILDCARD = os.path.join(DATA_PATH, 'sgpstamp*202001*.nc')
EXAMPLE_NOAA_PSL = os.path.join(DATA_PATH, 'ctd21125.15w')



"""
Module containing utilities for geographic calculations,
including solar calculations

"""

import numpy as np
import pandas as pd
from datetime import datetime, timezone, timedelta
from skyfield.api import wgs84, N, W, load_file, load
from skyfield import almanac
import re
import dateutil.parser
import pytz
from pathlib import Path
from act.utils.datetime_utils import datetime64_to_datetime
from act.utils.data_utils import convert_units

skyfield_bsp_file = str(Path(Path(__file__).parent, "conf", "de421.bsp"))


def destination_azimuth_distance(lat, lon, az, dist, dist_units='m'):
    """
    This procedure will calculate a destination lat/lon from
    an initial lat/lon and azimuth and distance.

    Parameters
    ----------
    lat : float
        Initial latitude.
    lon : float
        Initial longitude.
    az : float
        Azimuth in degrees.
    dist : float
        Distance
    dist_units : str
        Units for dist

    Returns
    -------
    lat2 : float
        Latitude of new point in degrees
    lon2 : float
        Longitude of new point in degrees

    """
    # Volumetric Mean Radius of Earth in km
    R = 6378.

    # Convert az to radian
    brng = np.radians(az)

    # Convert distance to km
    d = convert_units(dist, dist_units, 'km')

    # Convert lat/lon to radians
    lat = np.radians(lat)
    lon = np.radians(lon)

    # Using great circle equations
    lat2 = np.arcsin(np.sin(lat) * np.cos(d / R) +
                     np.cos(lat) * np.sin(d / R) * np.cos(brng))
    lon2 = lon + np.arctan2(np.sin(brng) * np.sin(d / R) * np.cos(lat),
                            np.cos(d / R) - np.sin(lat) * np.sin(lat2))

    return np.degrees(lat2), np.degrees(lon2)


def add_solar_variable(obj, latitude=None, longitude=None, solar_angle=0., dawn_dusk=False):
    """
    Add variable to the object to denote night (0) or sun (1).  If dawk_dusk is True
    will also return dawn (2) and dusk (3).  If at a high latitude and there's sun, will
    label twilight as dawn; if dark{2}, will label twilight as dusk(3).

    Parameters
    ----------
    obj : xarray dataset
        ACT object
    latitude : str
        Latitude variable name, default will look for matching variables in object
    longitude : str
        Longitude variable name, default will look for matching variables in object
    solar_angle : float
        Number of degress to use for dawn/dusk calculations
    dawn_dusk : boolean
         If set to True, will add values 2 (dawn) and 3 (dusk) to the solar variable

    Returns
    -------
    obj : xarray dataset
        Xarray object
    """

    variables = list(obj.keys())

    # Get coordinate variables
    if latitude is None:
        latitude = [s for s in variables if "latitude" in s]
        if len(latitude) == 0:
            latitude = [s for s in variables if "lat" in s]
        if len(latitude) == 0:
            raise ValueError("Latitude variable not set and could not be discerned from the data")

    if longitude is None:
        longitude = [s for s in variables if "longitude" in s]
        if len(longitude) == 0:
            longitude = [s for s in variables if "lon" in s]
        if len(longitude) == 0:
            raise ValueError("Longitude variable not set and could not be discerned from the data")

    # Get lat/lon variables
    lat = obj[latitude[0]].values
    lon = obj[longitude[0]].values

    # Loop through each time to ensure that the sunrise/set calcuations
    # are correct for each time and lat/lon if multiple
    results = is_sun_visible(latitude=lat, longitude=lon, date_time=obj['time'].values,
                             dawn_dusk=dawn_dusk)

    # Set longname
    longname = 'Daylight indicator; 0-Night; 1-Sun'

    if dawn_dusk is False:
        results = results * 1
    else:
        # If dawn_dusk is True, add 2 more indicators
        longname += '; 2-Dawn; 3-Dusk; 4-Twilight'
        dark_ind = np.where((results == 0))[0]
        twil_ind = np.where((results > 0) & (results < 4))[0]
        sun_ind = np.where((results == 4))[0]

        if len(sun_ind) == 0:
            results[twil_ind] = 3
        elif len(dark_ind) == 0:
            results[twil_ind] = 2
            results[sun_ind] = 1
        else:
            # Set Dawn between dark and sun
            if dark_ind[-1] < sun_ind[0]:
                dawn_ind = list(range(dark_ind[-1], sun_ind[0]))
            else:
                dawn_ind = list(range(dark_ind[-1], len(results))) + list(range(0, sun_ind[0]))
            results[dawn_ind] = 2

            # Set Dusk between sun and dark
            if sun_ind[-1] < dark_ind[0]:
                dusk_ind = list(range(sun_ind[-1], dark_ind[0]))
            else:
                dusk_ind = list(range(sun_ind[-1], len(results))) + list(range(0, dark_ind[0]))
            results[dusk_ind] = 3
            results[sun_ind] = 1

    # Add results to object and return
    obj['sun_variable'] = ('time', np.array(results),
                           {'long_name': longname, 'units': ' '})

    return obj


def get_solar_azimuth_elevation(latitude=None, longitude=None, time=None, library='skyfield',
                                temperature_C='standard', pressure_mbar='standard'):
    """
    Calculate solar azimuth, elevation and solar distance.


    Parameters
    ----------
    latitude : int, float
        Latitude in degrees north positive. Must be a scalar.
    longitude : int, float
        Longitude in degrees east positive. Must be a scalar.
    time : datetime.datetime, numpy.datetime64, list, numpy.array
        Time in UTC. May be a scalar or vector. datetime must be timezone aware.
    library : str
        Library to use for making calculations. Options include ['skyfield']
    temperature_C : string or list of float
        If library is 'skyfield' the temperature in degrees C at the surface for
        atmospheric compensation of the positon of the sun. Set to None for no
        compensation or 'standard' for standard model with a standard temperature.
    pressure_mbar : string or list of float
        If library is 'skyfield' the pressure in milibars at the surface for
        atmospheric compensation of the positon of the sun. Set to None for no
        compensation or 'standard' for standard model with a standard pressure.

    Returns
    -------
    result : tuple of float
        Values returned are a tuple of elevation, azimuth and distance. Elevation and
        azimuth are in degrees, with distance in Astronomical Units.

    """

    result = {'elevation': None, 'azimuth': None, 'distance': None}

    if library == 'skyfield':
        planets = load_file(skyfield_bsp_file)
        earth, sun = planets['earth'], planets['sun']

        if isinstance(time, datetime) and time.tzinfo is None:
            time = time.replace(tzinfo=pytz.UTC)

        if isinstance(time, (list, tuple)) and time[0].tzinfo is None:
            time = [ii.replace(tzinfo=pytz.UTC) for ii in time]

        if type(time).__module__ == np.__name__ and np.issubdtype(time.dtype, np.datetime64):
            time = time.astype('datetime64[s]').astype(int)
            if time.size > 1:
                time = [datetime.fromtimestamp(tm, timezone.utc) for tm in time]
            else:
                time = [datetime.fromtimestamp(time, timezone.utc)]

        if not isinstance(time, (list, tuple, np.ndarray)):
            time = [time]

        ts = load.timescale()
        t = ts.from_datetimes(time)
        location = earth + wgs84.latlon(latitude * N, longitude * W)
        astrometric = location.at(t).observe(sun)
        alt, az, distance = astrometric.apparent().altaz(temperature_C=temperature_C,
                                                         pressure_mbar=pressure_mbar)
        result = (alt.degrees, az.degrees, distance.au)
        planets.close()

    return result


def get_sunrise_sunset_noon(latitude=None, longitude=None, date=None, library='skyfield',
                            timezone=False):
    """
    Calculate sunrise, sunset and local solar noon times.

    Parameters
    ----------
    latitude : int, float
        Latitude in degrees north positive. Must be a scalar.
    longitude : int, float
        Longitude in degrees east positive. Must be a scalar.
    date : (datetime.datetime, numpy.datetime64, list of datetime.datetime,
            numpy.array of numpy.datetime64, string, list of string)
        Date(s) to return sunrise, sunset and noon times spaning the first date to last
        date if more than one provided. May be a scalar or vector. If entered as a string must follow
        YYYYMMDD format.
    library : str
        Library to use for making calculations. Options include ['skyfield']
    timezone : boolean
        Have timezone with datetime.

    Returns
    -------
    result : tuple of three numpy.array
        Tuple of three values sunrise, sunset, noon. Values will be a list.
        If no values can be calculated will return empty list. If the date is within
        polar night will return empty lists. If spans the transition to polar day
        will return previous sunrise or next sunset outside of date range provided.
    """

    sunrise, sunset, noon = np.array([]), np.array([]), np.array([])

    if library == 'skyfield':
        ts = load.timescale()
        eph = load_file(skyfield_bsp_file)
        sf_dates = []

        # Parse datetime object
        if isinstance(date, datetime):
            if date.tzinfo is None:
                sf_dates = [date.replace(tzinfo=pytz.UTC)]
            else:
                sf_dates = [date]

        if isinstance(date, (list, tuple)) and isinstance(date[0], datetime):
            if date[0].tzinfo is not None:
                sf_dates = date
            else:
                sf_dates = [ii.replace(tzinfo=pytz.UTC) for ii in date]

        # Parse string date
        if isinstance(date, str):
            sf_dates = [datetime.strptime(date, '%Y%m%d').replace(tzinfo=pytz.UTC)]

        # Parse list of string dates
        if isinstance(date, (list, tuple)) and isinstance(date[0], str):
            sf_dates = [datetime.strptime(dt, '%Y%m%d').replace(tzinfo=pytz.UTC) for dt in date]

        # Convert datetime64 to datetime
        if type(date).__module__ == np.__name__ and np.issubdtype(date.dtype, np.datetime64):
            sf_dates = datetime64_to_datetime(date)
            sf_dates = [ii.replace(tzinfo=pytz.UTC) for ii in sf_dates]

        # Function for calculating solar noon
        # Convert location into skyfield location object
        location = wgs84.latlon(latitude, longitude)
        # Set up function to indicate calculating locatin of Sun from Earth
        f = almanac.meridian_transits(eph, eph['Sun'], location)
        # Set up dates to be start of day and end of day so have a range
        t0 = sf_dates[0]
        t0 = t0.replace(hour=0, minute=0, second=0)
        t1 = sf_dates[-1]
        t1 = t1.replace(hour=23, minute=59, second=59)
        # Convert times from datetime to skyfild times
        t0 = ts.from_datetime(t0)
        t1 = ts.from_datetime(t1)
        # Calculate Meridian Transits. n contains times and x contains 1 and 0's
        # indicating when transit time is above or below location.
        n, x = almanac.find_discrete(t0, t1, f)

        # Determine if time is during daylight
        f = almanac.sunrise_sunset(eph, location)
        sun_up = f(n)

        # Filter out times when sun is below location or in polar night
        n = n[(x == 1) & sun_up]
        noon = n.utc_datetime()
        if noon.size == 0:
            return sunrise, sunset, noon

        # Calcuate sunrise and sunset times. Calcuate over range 12 less than minimum
        # noon time and 12 hours greater than maximum noon time.
        t0 = min(noon) - timedelta(hours=12)
        t1 = max(noon) + timedelta(hours=12)
        t0 = ts.from_datetime(t0)
        t1 = ts.from_datetime(t1)
        f = almanac.sunrise_sunset(eph, location)
        t, y = almanac.find_discrete(t0, t1, f)
        times = t.utc_datetime()
        sunrise = times[y == 1]
        sunset = times[y == 0]

        # Fill in sunrise and sunset if asked to during polar day
        if len(noon) > 0 and (y.size == 0 or len(sunrise) != len(sunset)):
            days = 200
            t0 = min(noon) - timedelta(days=days)
            t1 = max(noon) + timedelta(days=days)
            t0 = ts.from_datetime(t0)
            t1 = ts.from_datetime(t1)
            t, yy = almanac.find_discrete(t0, t1, f)
            times = t.utc_datetime()

            # If first time is sunset and/or last time is sunrise filter
            # from times
            if yy[0] == 0:
                yy = yy[1:]
                times = times[1:]
            if yy[-1] == 1:
                yy = yy[:-1]
                times = times[:-1]

            # Extract sunrise times
            temp_sunrise = times[yy == 1]
            # Extract sunset times
            temp_sunset = times[yy == 0]
            # Look for index closest to first noon time to get the time of last sunrise
            # since we are in polar day.
            diff = temp_sunrise - min(noon)
            sunrise_index = np.max(np.where(diff < timedelta(seconds=1)))
            # Look for index closest to last noon time to get the time of first sunset
            # since we are in polar day.
            diff = max(noon) - temp_sunset
            sunset_index = np.min(np.where(diff < timedelta(seconds=1))) + 1
            sunrise = temp_sunrise[sunrise_index: sunset_index]
            sunset = temp_sunset[sunrise_index: sunset_index]

        eph.close()

    if timezone is False:
        for ii in range(0, sunset.size):
            sunrise[ii] = sunrise[ii].replace(tzinfo=None)
            sunset[ii] = sunset[ii].replace(tzinfo=None)
        for ii in range(0, noon.size):
            noon[ii] = noon[ii].replace(tzinfo=None)

    return sunrise, sunset, noon


def is_sun_visible(latitude=None, longitude=None, date_time=None, dawn_dusk=False):
    """
    Determine if sun is above horizon at for a list of times.

    Parameters
    ----------
    latitude : int, float
        Latitude in degrees north positive. Must be a scalar.
    longitude : int, float
        Longitude in degrees east positive. Must be a scalar.
    date_time : datetime.datetime, numpy.array.datetime64, list of datetime.datetime
        Datetime with timezone, datetime with no timezone in UTC, or numpy.datetime64
        format in UTC. Can be a single datetime object or list of datetime objects.
    dawn_dusk : boolean
        If set to True, will use skyfields dark_twilight_day function to calculate sun up
        Returns a list of int's instead of boolean.
        0 - Dark of Night
        1 - Astronomical Twilight
        2 - Nautical Twilight
        3 - Civil Twilight
        4 - Sun Is Up

    Returns
    -------
    result : list
        List matching size of date_time containing True/False if sun is above horizon.
    """

    sf_dates = None

    # Check if datetime object is scalar and if has no timezone.
    if isinstance(date_time, datetime):
        if date_time.tzinfo is None:
            sf_dates = [date_time.replace(tzinfo=pytz.UTC)]
        else:
            sf_dates = [date_time]

    # Check if datetime objects in list have timezone. If not add.
    if isinstance(date_time, (list, tuple)) and isinstance(date_time[0], datetime):
        if isinstance(date_time[0], datetime) and date_time[0].tzinfo is not None:
            sf_dates = date_time
        else:
            sf_dates = [ii.replace(tzinfo=pytz.UTC) for ii in date_time]

    # Convert datetime64 to datetime with timezone.
    if type(date_time).__module__ == np.__name__ and np.issubdtype(date_time.dtype, np.datetime64):
        sf_dates = datetime64_to_datetime(date_time)
        sf_dates = [ii.replace(tzinfo=pytz.UTC) for ii in sf_dates]

    if sf_dates is None:
        raise ValueError('The date_time values entered into is_sun_visible() '
                         'do not match input types.')

    ts = load.timescale()
    eph = load_file(skyfield_bsp_file)

    t0 = ts.from_datetimes(sf_dates)
    location = wgs84.latlon(latitude, longitude)
    if dawn_dusk:
        f = almanac.dark_twilight_day(eph, location)
    else:
        f = almanac.sunrise_sunset(eph, location)

    sun_up = f(t0)

    eph.close()

    return sun_up



"""
This module contains functions for correcting raman lidar data

"""
import numpy as np


def correct_rl(obj, var_name='depolarization_counts_high', fill_value=1e-7,
               range_normalize_log_values=False):
    """
    This procedure corrects raman lidar data by filling all zero and
    negative values of backscatter with fill_value and then converting
    the backscatter data into logarithmic space. It will also look for
    a coordinate height dimension, and if one is not found will create
    using values from global attributes.

    Parameters
    ----------
    obj : Dataset object
        The doppler lidar dataset to correct. The backscatter data should be
        in linear space.
    var_name : str
        The variable name of data in the object.
    fill_value : float
        The fill_value to use. The fill_value is entered in linear space.
    range_normalize_log_values : boolean
        Option to range normalize and convert to log scale of counts values.

    Returns
    -------
    obj : Dataset object
        The raman lidar dataset containing the corrected values.

    """
    # This will get the name of the coordinate dimension so it's not assumed
    # via position or name.
    height_name = list(set(obj[var_name].dims) - set(['time']))[0]

    # Check if the height dimension is a variable in the object. If not
    # use global attributes to derive the values and put into object.
    #
    # Asking for a variable name in the object that is also a dimension
    # but does not exist will return an index array starting at 0.
    height = obj[height_name].values
    if height_name not in list(obj.data_vars):
        # Determin which mode we are correcting
        level = height_name.split('_')[0]
        att_name = [i for i in list(obj.attrs) if
                    'vertical_resolution_' + level + '_channels' in i]

        # Extract information from global attributes
        bin_size_raw = (obj.attrs[att_name[0]]).split()
        bins_before_shot = float(obj.attrs['number_of_bins_before_shot'])
        bin_size = float(bin_size_raw[0])
        height = (height + 1) * bin_size
        height = height - bins_before_shot * bin_size
        obj[height_name] = (height_name, height,
                            {'long_name': 'Height above ground',
                             'units': bin_size_raw[1]})

    if range_normalize_log_values:
        height = height ** 2  # Range normalize values
        backscat = obj[var_name].values
        # Doing this trick with height to change the array shape so it
        # will broadcast correclty against backscat
        backscat = backscat * height[None, :]
        backscat[backscat <= 0] = fill_value
        if np.shape(obj[var_name].values) != np.shape(np.log10(backscat)):
            obj[var_name].values = np.reshape(np.log10(backscat),
                                              np.shape(obj[var_name].values))
        else:
            obj[var_name].values = np.log10(backscat)

        # Updating the units to correctly indicate the values are log values
        obj[var_name].attrs['units'] = 'log(' + obj[var_name].attrs['units'] + ')'

    return obj



import act
from act.io.noaagml import read_gml
import act.tests.sample_files as sample_files
from pathlib import Path
import tempfile
import numpy as np
import glob


def test_io():
    sonde_ds = act.io.armfiles.read_netcdf([act.tests.EXAMPLE_MET1])
    assert 'temp_mean' in sonde_ds.variables.keys()
    assert 'rh_mean' in sonde_ds.variables.keys()
    assert sonde_ds.attrs['_arm_standards_flag'] == (1 << 0)

    with np.testing.assert_raises(OSError):
        act.io.armfiles.read_netcdf([])

    result = act.io.armfiles.read_netcdf([], return_None=True)
    assert result is None
    result = act.io.armfiles.read_netcdf(['./randomfile.nc'], return_None=True)
    assert result is None

    obj = act.io.armfiles.read_netcdf([act.tests.EXAMPLE_MET_TEST1])
    assert 'time' in obj

    obj = act.io.armfiles.read_netcdf([act.tests.EXAMPLE_MET_TEST2])
    assert obj['time'].values[10] == np.datetime64('2019-01-01T00:10:00')
    sonde_ds.close()


def test_io_mfdataset():
    sonde_ds = act.io.armfiles.read_netcdf(
        act.tests.EXAMPLE_MET_WILDCARD)
    assert 'temp_mean' in sonde_ds.variables.keys()
    assert 'rh_mean' in sonde_ds.variables.keys()
    assert len(sonde_ds.attrs['_file_times']) == 7
    assert sonde_ds.attrs['_arm_standards_flag'] == (1 << 0)
    sonde_ds.close()


def test_io_csv():
    headers = ['day', 'month', 'year', 'time', 'pasquill',
               'wdir_60m', 'wspd_60m', 'wdir_60m_std',
               'temp_60m', 'wdir_10m', 'wspd_10m',
               'wdir_10m_std', 'temp_10m', 'temp_dp',
               'rh', 'avg_temp_diff', 'total_precip',
               'solar_rad', 'net_rad', 'atmos_press',
               'wv_pressure', 'temp_soil_10cm',
               'temp_soil_100cm', 'temp_soil_10ft']
    anl_ds = act.io.csvfiles.read_csv(
        act.tests.EXAMPLE_ANL_CSV, sep='\s+', column_names=headers)
    assert 'temp_60m' in anl_ds.variables.keys()
    assert 'rh' in anl_ds.variables.keys()
    assert anl_ds['temp_60m'].values[10] == -1.7
    anl_ds.close()

    files = glob.glob(act.tests.EXAMPLE_MET_CSV)
    obj = act.io.csvfiles.read_csv(files[0])
    assert 'date_time' in obj
    assert '_datastream' in obj.attrs


def test_io_dod():
    dims = {'time': 1440, 'drop_diameter': 50}

    try:
        obj = act.io.armfiles.create_obj_from_arm_dod('vdis.b1', dims, version='1.2',
                                                      scalar_fill_dim='time')
        assert 'moment1' in obj
        assert len(obj['base_time'].values) == 1440
        assert len(obj['drop_diameter'].values) == 50
        with np.testing.assert_warns(UserWarning):
            obj2 = act.io.armfiles.create_obj_from_arm_dod('vdis.b1', dims,
                                                           scalar_fill_dim='time')
        assert 'moment1' in obj2
        assert len(obj2['base_time'].values) == 1440
        assert len(obj2['drop_diameter'].values) == 50
        with np.testing.assert_raises(ValueError):
            obj = act.io.armfiles.create_obj_from_arm_dod('vdis.b1', {}, version='1.2')

    except Exception:
        return
    obj.close()
    obj2.close()


def test_io_write():
    sonde_ds = act.io.armfiles.read_netcdf(sample_files.EXAMPLE_SONDE1)
    sonde_ds.clean.cleanup()

    with tempfile.TemporaryDirectory() as tmpdirname:
        write_file = Path(tmpdirname, Path(sample_files.EXAMPLE_SONDE1).name)
        keep_vars = ['tdry', 'qc_tdry', 'dp', 'qc_dp']
        for var_name in list(sonde_ds.data_vars):
            if var_name not in keep_vars:
                del sonde_ds[var_name]
        sonde_ds.write.write_netcdf(path=write_file, FillValue=-9999)

        sonde_ds_read = act.io.armfiles.read_netcdf(str(write_file))
        assert list(sonde_ds_read.data_vars) == keep_vars
        assert isinstance(sonde_ds_read['qc_tdry'].attrs['flag_meanings'], str)
        assert sonde_ds_read['qc_tdry'].attrs['flag_meanings'].count('__') == 21
        for attr in ['qc_standards_version', 'qc_method', 'qc_comment']:
            assert attr not in list(sonde_ds_read.attrs)
        sonde_ds_read.close()
        del sonde_ds_read

    sonde_ds.close()

    sonde_ds = act.io.armfiles.read_netcdf(sample_files.EXAMPLE_EBBR1)
    sonde_ds.clean.cleanup()
    assert 'fail_min' in sonde_ds['qc_home_signal_15'].attrs
    assert 'standard_name' in sonde_ds['qc_home_signal_15'].attrs
    assert 'flag_masks' in sonde_ds['qc_home_signal_15'].attrs

    with tempfile.TemporaryDirectory() as tmpdirname:
        cf_convention = 'CF-1.8'
        write_file = Path(tmpdirname, Path(sample_files.EXAMPLE_EBBR1).name)
        sonde_ds.write.write_netcdf(path=write_file, make_copy=False, join_char='_',
                                    cf_compliant=True, cf_convention=cf_convention)

        sonde_ds_read = act.io.armfiles.read_netcdf(str(write_file))

        assert cf_convention in sonde_ds_read.attrs['Conventions'].split()
        assert sonde_ds_read.attrs['FeatureType'] == 'timeSeries'
        global_att_keys = [ii for ii in sonde_ds_read.attrs.keys() if not ii.startswith('_')]
        assert global_att_keys[-1] == 'history'
        assert sonde_ds_read['alt'].attrs['axis'] == 'Z'
        assert sonde_ds_read['alt'].attrs['positive'] == 'up'

        sonde_ds_read.close()
        del sonde_ds_read

    sonde_ds.close()

    obj = act.io.armfiles.read_netcdf(sample_files.EXAMPLE_CEIL1)
    with tempfile.TemporaryDirectory() as tmpdirname:
        cf_convention = 'CF-1.8'
        write_file = Path(tmpdirname, Path(sample_files.EXAMPLE_CEIL1).name)
        obj.write.write_netcdf(path=write_file, make_copy=False, join_char='_',
                               cf_compliant=True, cf_convention=cf_convention)

        obj_read = act.io.armfiles.read_netcdf(str(write_file))

        assert cf_convention in obj_read.attrs['Conventions'].split()
        assert obj_read.attrs['FeatureType'] == 'timeSeriesProfile'
        assert len(obj_read.dims) > 1

        obj_read.close()
        del obj_read


def test_io_mpldataset():
    try:
        mpl_ds = act.io.mpl.read_sigma_mplv5(
            act.tests.EXAMPLE_SIGMA_MPLV5)
    except Exception:
        return

    # Tests fields
    assert 'channel_1' in mpl_ds.variables.keys()
    assert 'temp_0' in mpl_ds.variables.keys()
    assert mpl_ds.channel_1.values.shape == (102, 1000)

    # Tests coordinates
    assert 'time' in mpl_ds.coords.keys()
    assert 'range' in mpl_ds.coords.keys()
    assert mpl_ds.coords['time'].values.shape == (102, )
    assert mpl_ds.coords['range'].values.shape == (1000, )
    assert '_arm_standards_flag' in mpl_ds.attrs.keys()

    # Tests attributes
    assert '_datastream' in mpl_ds.attrs.keys()
    mpl_ds.close()


def test_read_gml():
    # Test Radiation
    ds = read_gml(sample_files.EXAMPLE_GML_RADIATION, datatype='RADIATION')
    assert np.isclose(np.nansum(ds['solar_zenith_angle']), 1629.68)
    assert np.isclose(np.nansum(ds['upwelling_infrared_case_temp']), 4185.73)
    assert (ds['upwelling_infrared_case_temp'].attrs['ancillary_variables'] ==
            'qc_upwelling_infrared_case_temp')
    assert ds['qc_upwelling_infrared_case_temp'].attrs['flag_values'] == [0, 1, 2]
    assert (ds['qc_upwelling_infrared_case_temp'].attrs['flag_meanings'] ==
            ['Not failing any tests', 'Knowingly bad value', 'Should be used with scrutiny'])
    assert (ds['qc_upwelling_infrared_case_temp'].attrs['flag_assessments'] ==
            ['Good', 'Bad', 'Indeterminate'])
    assert ds['time'].values[-1] == np.datetime64('2021-01-01T00:17:00')

    ds = read_gml(sample_files.EXAMPLE_GML_RADIATION, convert_missing=False)
    assert np.isclose(np.nansum(ds['solar_zenith_angle']), 1629.68)
    assert np.isclose(np.nansum(ds['upwelling_infrared_case_temp']), 4185.73)
    assert (ds['upwelling_infrared_case_temp'].attrs['ancillary_variables'] ==
            'qc_upwelling_infrared_case_temp')
    assert ds['qc_upwelling_infrared_case_temp'].attrs['flag_values'] == [0, 1, 2]
    assert (ds['qc_upwelling_infrared_case_temp'].attrs['flag_meanings'] ==
            ['Not failing any tests', 'Knowingly bad value', 'Should be used with scrutiny'])
    assert (ds['qc_upwelling_infrared_case_temp'].attrs['flag_assessments'] ==
            ['Good', 'Bad', 'Indeterminate'])
    assert ds['time'].values[-1] == np.datetime64('2021-01-01T00:17:00')

    # Test MET
    ds = read_gml(sample_files.EXAMPLE_GML_MET, datatype='MET')
    assert np.isclose(np.nansum(ds['wind_speed'].values), 140.999)
    assert ds['wind_speed'].attrs['units'] == 'm/s'
    assert np.isnan(ds['wind_speed'].attrs['_FillValue'])
    assert np.sum(np.isnan(ds['preciptation_intensity'].values)) == 19
    assert ds['preciptation_intensity'].attrs['units'] == 'mm/hour'
    assert ds['time'].values[0] == np.datetime64('2020-01-01T01:00:00')

    ds = read_gml(sample_files.EXAMPLE_GML_MET, convert_missing=False)
    assert np.isclose(np.nansum(ds['wind_speed'].values), 140.999)
    assert ds['wind_speed'].attrs['units'] == 'm/s'
    assert np.isclose(ds['wind_speed'].attrs['_FillValue'], -999.9)
    assert np.sum(ds['preciptation_intensity'].values) == -1881
    assert ds['preciptation_intensity'].attrs['units'] == 'mm/hour'
    assert ds['time'].values[0] == np.datetime64('2020-01-01T01:00:00')

    # Test Ozone
    ds = read_gml(sample_files.EXAMPLE_GML_OZONE, datatype='OZONE')
    assert np.isclose(np.nansum(ds['ozone'].values), 582.76)
    assert ds['ozone'].attrs['long_name'] == 'Ozone'
    assert ds['ozone'].attrs['units'] == 'ppb'
    assert np.isnan(ds['ozone'].attrs['_FillValue'])
    assert ds['time'].values[0] == np.datetime64('2020-12-01T00:00:00')

    ds = read_gml(sample_files.EXAMPLE_GML_OZONE)
    assert np.isclose(np.nansum(ds['ozone'].values), 582.76)
    assert ds['ozone'].attrs['long_name'] == 'Ozone'
    assert ds['ozone'].attrs['units'] == 'ppb'
    assert np.isnan(ds['ozone'].attrs['_FillValue'])
    assert ds['time'].values[0] == np.datetime64('2020-12-01T00:00:00')

    # Test Carbon Dioxide
    ds = read_gml(sample_files.EXAMPLE_GML_CO2, datatype='co2')
    assert np.isclose(np.nansum(ds['co2'].values), 2307.630)
    assert (ds['qc_co2'].values ==
            np.array([1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0], dtype=int)).all()
    assert ds['co2'].attrs['units'] == 'ppm'
    assert np.isnan(ds['co2'].attrs['_FillValue'])
    assert ds['qc_co2'].attrs['flag_assessments'] == ['Bad', 'Indeterminate']
    assert ds['latitude'].attrs['standard_name'] == 'latitude'

    ds = read_gml(sample_files.EXAMPLE_GML_CO2, convert_missing=False)
    assert np.isclose(np.nansum(ds['co2'].values), -3692.3098)
    assert ds['co2'].attrs['_FillValue'] == -999.99
    assert (ds['qc_co2'].values ==
            np.array([1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0], dtype=int)).all()
    assert ds['co2'].attrs['units'] == 'ppm'
    assert np.isclose(ds['co2'].attrs['_FillValue'], -999.99)
    assert ds['qc_co2'].attrs['flag_assessments'] == ['Bad', 'Indeterminate']
    assert ds['latitude'].attrs['standard_name'] == 'latitude'

    # Test Halocarbon
    ds = read_gml(sample_files.EXAMPLE_GML_HALO, datatype='HALO')
    assert np.isclose(np.nansum(ds['CCl4'].values), 1342.6499)
    assert ds['CCl4'].attrs['units'] == 'ppt'
    assert ds['CCl4'].attrs['long_name'] == 'Carbon Tetrachloride (CCl4) daily median'
    assert np.isnan(ds['CCl4'].attrs['_FillValue'])
    assert ds['time'].values[0] == np.datetime64('1998-06-16T00:00:00')

    ds = read_gml(sample_files.EXAMPLE_GML_HALO)
    assert np.isclose(np.nansum(ds['CCl4'].values), 1342.6499)
    assert ds['CCl4'].attrs['units'] == 'ppt'
    assert ds['CCl4'].attrs['long_name'] == 'Carbon Tetrachloride (CCl4) daily median'
    assert np.isnan(ds['CCl4'].attrs['_FillValue'])
    assert ds['time'].values[0] == np.datetime64('1998-06-16T00:00:00')


def test_read_psl_wind_profiler():
    test_obj_low, test_obj_hi = act.io.noaapsl.read_psl_wind_profiler(
        act.tests.EXAMPLE_NOAA_PSL, transpose=False)
    # test dimensions
    assert 'time' and 'height' in test_obj_low.dims.keys()
    assert 'time' and 'height' in test_obj_hi.dims.keys()
    assert test_obj_low.dims['time'] == 4
    assert test_obj_hi.dims['time'] == 3
    assert test_obj_low.dims['height'] == 49
    assert test_obj_hi.dims['height'] == 50

    # test coordinates
    assert (test_obj_low.coords['height'][0:5] == np.array(
        [0.151, 0.254, 0.356, 0.458, 0.561])).all()
    assert (test_obj_low.coords['time'][0:2] == np.array([
        '2021-05-05T15:00:01.000000000', '2021-05-05T15:15:49.000000000'],
        dtype='datetime64[ns]')).all()

    # test attributes
    assert test_obj_low.attrs['site_identifier'] == 'CTD'
    assert test_obj_low.attrs['data_type'] == 'WINDS'
    assert test_obj_low.attrs['revision_number'] == 'rev 5.1'
    assert test_obj_low.attrs['latitude'] == 34.66
    assert test_obj_low.attrs['longitude'] == -87.35
    assert test_obj_low.attrs['altitude'] == 187.0
    assert (test_obj_low.attrs['azimuth'] == np.array(
        [38., 38., 308.], dtype='float32')).all()
    assert (test_obj_low.attrs['elevation'] == np.array(
        [90., 74.7, 74.7], dtype='float32')).all()

    # test fields
    assert test_obj_low['RAD1'].shape == (4, 49)
    assert test_obj_hi['RAD1'].shape == (3, 50)
    assert (test_obj_low['RAD1'][0, 0:5] == np.array(
        [0.2, 0.1, 0.1, 0., -0.1])).all()
    assert (test_obj_hi['RAD1'][0, 0:5] == np.array(
        [0.1, 0.1, -0.1, 0., -0.2])).all()

    assert test_obj_low['SPD'].shape == (4, 49)
    assert test_obj_hi['SPD'].shape == (3, 50)
    assert (test_obj_low['SPD'][0, 0:5] == np.array(
        [2.5, 3.3, 4.3, 4.3, 4.8])).all()
    assert (test_obj_hi['SPD'][0, 0:5] == np.array(
        [3.7, 4.6, 6.3, 5.2, 6.8])).all()

    # test transpose
    test_obj_low, test_obj_hi = act.io.noaapsl.read_psl_wind_profiler(
        act.tests.EXAMPLE_NOAA_PSL, transpose=True)
    assert test_obj_low['RAD1'].shape == (49, 4)
    assert test_obj_hi['RAD1'].shape == (50, 3)
    assert test_obj_low['SPD'].shape == (49, 4)
    assert test_obj_hi['SPD'].shape == (50, 3)
    test_obj_low.close()



"""
Functions for aeri retrievals.

"""

import numpy as np
from scipy.optimize import brentq
from act.retrievals.irt import irt_response_function, sum_function_irt


def aeri2irt(aeri_ds, wnum_name='wnum', rad_name='mean_rad', hatch_name='hatchOpen',
             tolerance=0.1, temp_low=150.0, temp_high=320.0, maxiter=200):
    """
    This function will integrate over the correct wavenumber values to produce
    the effective IRT temperature.

    As a note from the ARM IRT Instrument Handbook
    A positive bias of the sky temperature is exhibited by the downwelling IRT,
    compared to the AERI, during clear-sky conditions when the sky temperature
    is less than ~180K. The effect depends on the characteristics of the
    individual IRT and the internal reference temperature of the IRT. The
    greatest difference compared to AERI will occur when the sky is very clear,
    dry, and cold and the ambient temperature is relatively hot, maximizing the
    difference in temperature between the sky and instrument, and the
    calibration of the IRT at the lower limit of 223K was not performed
    accurately. This bias is especially apparent at high-latitude sites
    (e.g., NSA, OLI, and AWR).

    https://www.arm.gov/publications/tech_reports/handbooks/irt_handbook.pdf

    Author - Ken Kehoe

    Parameters
    ----------
    aeri_ds : Xarray Dataset Object
        The Dataset object containing AERI data.
    wnum_name : str
        The variable name for coordinate dimention of wave number Xarray Dataset.
    hatch_name : str or None
        The variable name for hatch status. If set to None will not try to set
        when hatch is not opent to NaN.
    rad_name : str
        The variable name for mean radiance in Xarray Dataset.
    tolerance : float
        The tolerance value to try and match for returned temperature.
    temp_low : float
        The initial low value to use in zbren function to invert radiances.
    temp_high : float
        The initial low value to use in zbren function to invert radiances.
    maxiter : int
        The maximum number if iterations to use with invertion process.
        Prevents runaway processes.

    Returns
    -------
    obj : Xarray Dataset Object or None
        The aeri_ds Dataset with new DataArray of temperatures added under
        variable name 'aeri_irt_equiv_temperature'.

    """
    # Get data values
    rf_wnum, rf = irt_response_function()
    wnum = aeri_ds[wnum_name].values
    mean_rad = aeri_ds[rad_name].values

    # Pull out AERI data for correct wavenumbers and apply response function --;
    index = np.where((wnum >= (rf_wnum[0] - 0.001)) & (wnum <= (rf_wnum[-1] + 0.001)))[0]
    if index.size == 0:
        raise ValueError('No wavenumbers match for aeri2irt')

    wnum = wnum[index]
    mean_rad = mean_rad[:, index]
    # If the wavenumbers in AERI data are not close enough to response function
    # match the wavenumbers and adjust.
    atol = 0.001
    if not np.all(np.isclose(wnum, rf_wnum, atol=atol)):
        index_wnum = []
        index_rf = []
        for ii in range(wnum.size):
            idx = (np.abs(wnum[ii] - rf_wnum)).argmin()
            if np.isclose(wnum[ii], rf_wnum[idx], atol=atol):
                index_wnum.append(ii)
                index_rf.append(idx)
        mean_rad = mean_rad[:, index_wnum]
        rf = rf[index_rf]

    # Apply response function to AERI radiance
    mean_rad = mean_rad * rf

    # Sum along wavenumber dimention to get a single value for each time step
    mean_rad = np.nansum(mean_rad, axis=1)

    # Loop over each time step of the AERI data and through the use of
    # solving for zero determine the AERI equivlante IRT sky temperature.
    aeri_irt_vals = np.full(mean_rad.size, np.nan, dtype=mean_rad.dtype)

    # Look for when the hatch is not in Open position and set values to NaN.
    if hatch_name is not None:
        flag_values = aeri_ds[hatch_name].attrs['flag_values']
        flag_meanings = aeri_ds[hatch_name].attrs['flag_meanings']
        if not isinstance(flag_meanings, list):
            flag_meanings = flag_meanings.split()
        flag_meanings = [att.lower() for att in flag_meanings]
        if not isinstance(flag_values, list):
            flag_values = flag_values.split()
            flag_values = [int(att) for att in flag_values]
        value = flag_values[flag_meanings.index('open')]
        mean_rad[aeri_ds[hatch_name].values != value] = np.nan

    for ii in range(mean_rad.size):
        if np.isnan(mean_rad[ii]):
            continue
        else:
            try:
                aeri_irt_vals[ii] = brentq(sum_function_irt, temp_low, temp_high,
                                           args=(mean_rad[ii], ), xtol=tolerance, maxiter=maxiter)
            except ValueError:
                pass

    # Add new values to Xarray Dataset
    aeri_ds['aeri_irt_equiv_temperature'] = (
        'time', aeri_irt_vals,
        {'long_name': 'Derived IRT equivalent temperatrues from AERI',
         'units': 'K'})

    return aeri_ds



"""
Stores the class for WindRoseDisplay.

"""

import matplotlib.pyplot as plt
import numpy as np
import warnings

from .plot import Display
# Import Local Libs
from ..utils import datetime_utils as dt_utils


class WindRoseDisplay(Display):
    """
    A class for handing wind rose plots.

    This is inherited from the :func:`act.plotting.Display`
    class and has therefore has the same attributes as that class.
    See :func:`act.plotting.Display`
    for more information. There are no additional attributes or parameters
    to this class.

    Examples
    --------
    To create a WindRoseDisplay object, simply do:

    .. code-block :: python

        sonde_ds = act.io.armfiles.read_netcdf('sonde_data.nc')
        WindDisplay = act.plotting.WindRoseDisplay(sonde_ds, figsize=(8,10))

    """
    def __init__(self, obj, subplot_shape=(1,), ds_name=None, **kwargs):
        super().__init__(obj, subplot_shape, ds_name,
                         subplot_kw=dict(projection='polar'), **kwargs)

    def set_thetarng(self, trng=(0., 360.), subplot_index=(0,)):
        """
        Sets the theta range of the wind rose plot.

        Parameters
        ----------
        trng : 2-tuple
            The range (in degrees).
        subplot_index : 2-tuple
            The index of the subplot to set the degree range of.

        """
        if self.axes is not None:
            self.axes[subplot_index].set_thetamin(trng[0])
            self.axes[subplot_index].set_thetamax(trng[1])
            self.trng = trng
        else:
            raise RuntimeError(("Axes must be initialized before" +
                                " changing limits!"))
        print(self.trng)

    def set_rrng(self, rrng, subplot_index=(0,)):
        """
        Sets the range of the radius of the wind rose plot.

        Parameters
        ----------
        rrng : 2-tuple
            The range for the plot radius (in %).
        subplot_index : 2-tuple
            The index of the subplot to set the radius range of.

        """
        if self.axes is not None:
            self.axes[subplot_index].set_rmin(rrng[0])
            self.axes[subplot_index].set_rmax(rrng[1])
            self.rrng = rrng
        else:
            raise RuntimeError(("Axes must be initialized before" +
                                " changing limits!"))

    def plot(self, dir_field, spd_field, dsname=None, subplot_index=(0,),
             cmap=None, set_title=None, num_dirs=20, spd_bins=None,
             tick_interval=3, legend_loc=0, legend_bbox=None, legend_title=None,
             calm_threshold=1.,
             **kwargs):
        """
        Makes the wind rose plot from the given dataset.

        Parameters
        ----------
        dir_field : str
            The name of the field representing the wind direction (in degrees).
        spd_field : str
            The name of the field representing the wind speed.
        dsname : str
            The name of the datastream to plot from. Set to None to
            let ACT automatically try to determine this.
        subplot_index : 2-tuple
            The index of the subplot to place the plot on.
        cmap : str or matplotlib colormap
            The name of the matplotlib colormap to use.
        set_title : str
            The title of the plot.
        num_dirs : int
            The number of directions to split the wind rose into.
        spd_bins : 1D array-like
            The bin boundaries to sort the wind speeds into.
        tick_interval : int
            The interval (in %) for the ticks on the radial axis.
        legend_loc : int
            Legend location using matplotlib legend code
        legend_bbox : tuple
            Legend bounding box coordinates
        legend_title : string
            Legend title
        calm_threshold : float
            Winds below this threshold are considered to be calm.
        **kwargs : keyword arguments
            Additional keyword arguments will be passed into :func:plt.bar

        Returns
        -------
        ax : matplotlib axis handle
            The matplotlib axis handle corresponding to the plot.

        """
        if dsname is None and len(self._obj.keys()) > 1:
            raise ValueError(("You must choose a datastream when there are 2 "
                              "or more datasets in the TimeSeriesDisplay "
                              "object."))
        elif dsname is None:
            dsname = list(self._obj.keys())[0]

        # Get data and dimensions
        dir_data = self._obj[dsname][dir_field].values
        spd_data = self._obj[dsname][spd_field].values

        # Get the current plotting axis, add day/night background and plot data
        if self.fig is None:
            self.fig = plt.figure()

        if self.axes is None:
            self.axes = np.array([plt.axes(projection='polar')])
            self.fig.add_axes(self.axes[0])

        if spd_bins is None:
            spd_bins = np.linspace(0, np.nanmax(spd_data), 10)

        # Make the bins so that 0 degrees N is in the center of the first bin
        # We need to wrap around

        deg_width = 360. / num_dirs
        dir_bins_mid = np.linspace(0., 360. - 3 * deg_width / 2., num_dirs)
        wind_hist = np.zeros((num_dirs, len(spd_bins) - 1))

        for i in range(num_dirs):
            if i == 0:
                with warnings.catch_warnings():
                    warnings.filterwarnings("ignore", "invalid value encountered in.*")
                    the_range = np.logical_or(dir_data < deg_width / 2.,
                                              dir_data > 360. - deg_width / 2.)
            else:
                with warnings.catch_warnings():
                    warnings.filterwarnings("ignore", "invalid value encountered in.*")
                    the_range = np.logical_and(
                        dir_data >= dir_bins_mid[i] - deg_width / 2,
                        dir_data <= dir_bins_mid[i] + deg_width / 2)
            hist, bins = np.histogram(spd_data[the_range], spd_bins)
            wind_hist[i] = hist

        wind_hist = wind_hist / np.sum(wind_hist) * 100
        mins = np.deg2rad(dir_bins_mid)

        # Do the first level
        if 'units' in self._obj[dsname][spd_field].attrs.keys():
            units = self._obj[dsname][spd_field].attrs['units']
        else:
            units = ''
        the_label = ("%3.1f" % spd_bins[0] +
                     '-' + "%3.1f" % spd_bins[1] + " " + units)
        our_cmap = plt.cm.get_cmap(cmap)
        our_colors = our_cmap(np.linspace(0, 1, len(spd_bins)))

        bars = [self.axes[subplot_index].bar(mins, wind_hist[:, 0],
                                             bottom=0,
                                             label=the_label,
                                             width=0.8 * np.deg2rad(deg_width),
                                             color=our_colors[0],
                                             **kwargs)]
        for i in range(1, len(spd_bins) - 1):
            the_label = ("%3.1f" % spd_bins[i] +
                         '-' + "%3.1f" % spd_bins[i + 1] + " " + units)
            # Changing the bottom to be a sum of the previous speeds so that
            # it positions it correctly - Adam Theisen
            bars.append(self.axes[subplot_index].bar(
                mins, wind_hist[:, i], label=the_label,
                bottom=np.sum(wind_hist[:, :i], axis=1), width=0.8 * np.deg2rad(deg_width),
                color=our_colors[i], **kwargs))
        self.axes[subplot_index].legend(loc=legend_loc, bbox_to_anchor=legend_bbox,
                                        title=legend_title)
        self.axes[subplot_index].set_theta_zero_location("N")
        self.axes[subplot_index].set_theta_direction(-1)

        # Add an annulus with text stating % of time calm
        pct_calm = np.sum(spd_data <= calm_threshold) / len(spd_data) * 100
        self.axes[subplot_index].set_rorigin(-2.5)
        self.axes[subplot_index].annotate("%3.2f%%\n calm" % pct_calm, xy=(0, -2.5), ha='center', va='center')

        # Set the ticks to be nice numbers
        tick_max = tick_interval * round(
            np.nanmax(np.cumsum(wind_hist, axis=1)) / tick_interval)
        rticks = np.arange(0, tick_max, tick_interval)
        rticklabels = [("%d" % x + '%') for x in rticks]
        self.axes[subplot_index].set_rticks(rticks)
        self.axes[subplot_index].set_yticklabels(rticklabels)

        # Set Title
        if set_title is None:
            set_title = ' '.join([dsname, 'on',
                                  dt_utils.numpy_to_arm_date(
                                      self._obj[dsname].time.values[0])])
        self.axes[subplot_index].set_title(set_title)
        return self.axes[subplot_index]



" Unit tests for the ACT retrievals module. """

import act
import numpy as np
import xarray as xr
import glob


def test_get_stability_indices():
    sonde_ds = act.io.armfiles.read_netcdf(
        act.tests.sample_files.EXAMPLE_SONDE1)

    try:
        sonde_ds = act.retrievals.calculate_stability_indicies(
            sonde_ds, temp_name="tdry", td_name="dp", p_name="pres",
            rh_name='rh')
        metpy = True
    except ImportError:
        metpy = False

    if metpy is True:
        np.testing.assert_allclose(
            sonde_ds["parcel_temperature"].values[0:5],
            [269.85, 269.745276, 269.678006, 269.622444, 269.572331], rtol=1e-5)
        assert sonde_ds["parcel_temperature"].attrs["units"] == "kelvin"
        np.testing.assert_almost_equal(
            sonde_ds["surface_based_cape"], 0.62, decimal=2)
        assert sonde_ds["surface_based_cape"].attrs["units"] == "J/kg"
        assert sonde_ds[
            "surface_based_cape"].attrs["long_name"] == "Surface-based CAPE"
        np.testing.assert_almost_equal(
            sonde_ds["surface_based_cin"], 0.000, decimal=3)
        assert sonde_ds["surface_based_cin"].attrs["units"] == "J/kg"
        assert sonde_ds[
            "surface_based_cin"].attrs["long_name"] == "Surface-based CIN"
        np.testing.assert_almost_equal(
            sonde_ds["most_unstable_cape"], 0.000, decimal=3)
        assert sonde_ds["most_unstable_cape"].attrs["units"] == "J/kg"
        assert sonde_ds[
            "most_unstable_cape"].attrs["long_name"] == "Most unstable CAPE"
        np.testing.assert_almost_equal(
            sonde_ds["most_unstable_cin"], 0.000, decimal=3)
        assert sonde_ds["most_unstable_cin"].attrs["units"] == "J/kg"
        assert sonde_ds[
            "most_unstable_cin"].attrs["long_name"] == "Most unstable CIN"

        np.testing.assert_almost_equal(
            sonde_ds["lifted_index"], 28.4, decimal=1)
        assert sonde_ds["lifted_index"].attrs["units"] == "kelvin"
        assert sonde_ds["lifted_index"].attrs["long_name"] == "Lifted index"
        np.testing.assert_equal(
            sonde_ds["level_of_free_convection"], np.array(np.nan))
        assert sonde_ds[
            "level_of_free_convection"].attrs["units"] == "hectopascal"
        assert sonde_ds[
            "level_of_free_convection"].attrs[
                "long_name"] == "Level of free convection"
        np.testing.assert_almost_equal(
            sonde_ds["lifted_condensation_level_temperature"],
            -8.07, decimal=2)
        assert sonde_ds[
            "lifted_condensation_level_temperature"].attrs[
                "units"] == "degree_Celsius"
        assert sonde_ds[
            "lifted_condensation_level_temperature"].attrs[
                "long_name"] == "Lifted condensation level temperature"
        np.testing.assert_almost_equal(
            sonde_ds["lifted_condensation_level_pressure"], 927.1, decimal=1)
        assert sonde_ds[
            "lifted_condensation_level_pressure"].attrs[
                "units"] == "hectopascal"
        assert sonde_ds[
            "lifted_condensation_level_pressure"].attrs[
                "long_name"] == "Lifted condensation level pressure"
    sonde_ds.close()


def test_generic_sobel_cbh():
    ceil = act.io.armfiles.read_netcdf(
        act.tests.sample_files.EXAMPLE_CEIL1)

    ceil = ceil.resample(time='1min').nearest()
    ceil = act.retrievals.cbh.generic_sobel_cbh(
        ceil, variable='backscatter', height_dim='range',
        var_thresh=1000., fill_na=0)
    cbh = ceil['cbh_sobel'].values
    assert cbh[500] == 615.
    assert cbh[1000] == 555.
    ceil.close()


def test_calculate_precipitable_water():
    sonde_ds = act.io.armfiles.read_netcdf(
        act.tests.sample_files.EXAMPLE_SONDE1)
    assert sonde_ds["tdry"].units == "C", "Temperature must be in Celsius"
    assert sonde_ds["rh"].units == "%", "Relative Humidity must be a percentage"
    assert sonde_ds["pres"].units == "hPa", "Pressure must be in hPa"
    pwv_data = act.retrievals.calculate_precipitable_water(
        sonde_ds, temp_name='tdry', rh_name='rh', pres_name='pres')
    np.testing.assert_almost_equal(pwv_data, 0.8028, decimal=3)
    sonde_ds.close()


def test_doppler_lidar_winds():
    # Process a single file
    dl_ds = act.io.armfiles.read_netcdf(act.tests.sample_files.EXAMPLE_DLPPI)
    result = act.retrievals.doppler_lidar.compute_winds_from_ppi(
        dl_ds, intensity_name='intensity')
    assert np.round(
        np.nansum(result['wind_speed'].values)).astype(int) == 1570
    assert np.round(
        np.nansum(result['wind_direction'].values)).astype(int) == 32635
    assert result['wind_speed'].attrs['units'] == 'm/s'
    assert result['wind_direction'].attrs['units'] == 'degree'
    assert result['height'].attrs['units'] == 'm'
    dl_ds.close()

    # Process multiple files
    dl_ds = act.io.armfiles.read_netcdf(
        act.tests.sample_files.EXAMPLE_DLPPI_MULTI)
    del dl_ds['range'].attrs['units']
    result = act.retrievals.doppler_lidar.compute_winds_from_ppi(dl_ds)
    assert np.round(
        np.nansum(result['wind_speed'].values)).astype(int) == 64419
    assert np.round(
        np.nansum(result['wind_direction'].values)).astype(int) == 733627
    dl_ds.close()


def test_aeri2irt():
    aeri_ds = act.io.armfiles.read_netcdf(act.tests.sample_files.EXAMPLE_AERI)
    aeri_ds = act.retrievals.aeri.aeri2irt(aeri_ds)
    assert np.round(
        np.nansum(
            aeri_ds['aeri_irt_equiv_temperature'].values)).astype(int) == 17372
    np.testing.assert_almost_equal(
        aeri_ds['aeri_irt_equiv_temperature'].values[7], 286.081, decimal=3)
    np.testing.assert_almost_equal(
        aeri_ds['aeri_irt_equiv_temperature'].values[-10], 285.366, decimal=3)
    aeri_ds.close()
    del aeri_ds


def test_sst():
    obj = act.io.armfiles.read_netcdf(act.tests.sample_files.EXAMPLE_IRTSST)
    obj = act.retrievals.irt.sst_from_irt(obj)
    np.testing.assert_almost_equal(
        obj['sea_surface_temperature'].values[0], 278.901, decimal=3)
    np.testing.assert_almost_equal(
        obj['sea_surface_temperature'].values[-1], 279.291, decimal=3)
    assert np.round(
        np.nansum(obj['sea_surface_temperature'].values)).astype(int) == 6699
    obj.close()


def test_calculate_sirs_variable():
    sirs_object = act.io.armfiles.read_netcdf(
        act.tests.sample_files.EXAMPLE_SIRS)
    met_object = act.io.armfiles.read_netcdf(
        act.tests.sample_files.EXAMPLE_MET1)

    obj = act.retrievals.radiation.calculate_dsh_from_dsdh_sdn(sirs_object)
    assert np.isclose(np.nansum(obj['derived_down_short_hemisp'].values), 60350, atol=1)

    obj = act.retrievals.radiation.calculate_irradiance_stats(
        obj, variable='derived_down_short_hemisp',
        variable2='down_short_hemisp', threshold=60)
    assert np.isclose(np.nansum(obj['diff_derived_down_short_hemisp'].values), 527, atol=1)
    assert np.isclose(np.nansum(obj['ratio_derived_down_short_hemisp'].values), 392, atol=1)

    obj = act.retrievals.radiation.calculate_net_radiation(obj, smooth=30)
    assert np.ceil(np.nansum(obj['net_radiation'].values)) == 21915
    assert np.ceil(np.nansum(obj['net_radiation_smoothed'].values)) == 22316

    obj = act.retrievals.radiation.calculate_longwave_radiation(
        obj, temperature_var='temp_mean',
        vapor_pressure_var='vapor_pressure_mean',
        met_obj=met_object)
    assert np.ceil(obj['monteith_clear'].values[25]) == 239
    assert np.ceil(obj['monteith_cloudy'].values[30]) == 318
    assert np.ceil(obj['prata_clear'].values[35]) == 234

    new_obj = xr.merge([sirs_object, met_object], compat='override')
    obj = act.retrievals.radiation.calculate_longwave_radiation(
        new_obj, temperature_var='temp_mean',
        vapor_pressure_var='vapor_pressure_mean')
    assert np.ceil(obj['monteith_clear'].values[25]) == 239
    assert np.ceil(obj['monteith_cloudy'].values[30]) == 318
    assert np.ceil(obj['prata_clear'].values[35]) == 234

    sirs_object.close()
    met_object.close()


def test_calculate_pbl_liu_liang():
    files = glob.glob(act.tests.sample_files.EXAMPLE_TWP_SONDE_20060121)
    files2 = glob.glob(act.tests.sample_files.EXAMPLE_SONDE1)
    files += files2
    files.sort()

    pblht = []
    pbl_regime = []
    for i, r in enumerate(files):
        obj = act.io.armfiles.read_netcdf(r)
        obj['tdry'].attrs['units'] = 'degree_Celsius'
        obj = act.retrievals.sonde.calculate_pbl_liu_liang(obj, smooth_height=10)
        pblht.append(float(obj['pblht_liu_liang'].values))
        pbl_regime.append(obj['pblht_regime_liu_liang'].values)

    assert pbl_regime == ['NRL', 'NRL', 'NRL', 'NRL', 'NRL']
    np.testing.assert_array_almost_equal(pblht, [847.5, 858.2, 184.8, 197.7, 443.2], decimal=1)

    obj = act.io.armfiles.read_netcdf(files[1])
    obj['tdry'].attrs['units'] = 'degree_Celsius'
    obj = act.retrievals.sonde.calculate_pbl_liu_liang(obj, land_parameter=False)
    np.testing.assert_almost_equal(obj['pblht_liu_liang'].values, 733.6, decimal=1)

    obj = act.io.armfiles.read_netcdf(files[-2:])
    obj['tdry'].attrs['units'] = 'degree_Celsius'
    with np.testing.assert_raises(ValueError):
        obj = act.retrievals.sonde.calculate_pbl_liu_liang(obj)

    obj = act.io.armfiles.read_netcdf(files[0])
    obj['tdry'].attrs['units'] = 'degree_Celsius'
    temp = obj['tdry'].values
    temp[10:20] = 19.
    temp[0:10] = -10
    obj['tdry'].values = temp
    obj = act.retrievals.sonde.calculate_pbl_liu_liang(obj, land_parameter=False)
    assert obj['pblht_regime_liu_liang'].values == 'SBL'

    with np.testing.assert_raises(ValueError):
        obj2 = obj.where(obj['alt'] < 1000., drop=True)
        obj2 = act.retrievals.sonde.calculate_pbl_liu_liang(obj2, smooth_height=15)

    with np.testing.assert_raises(ValueError):
        obj2 = obj.where(obj['pres'] < 200., drop=True)
        obj2 = act.retrievals.sonde.calculate_pbl_liu_liang(obj2, smooth_height=15)

    with np.testing.assert_raises(ValueError):
        temp[0:5] = -40
        obj['tdry'].values = temp
        obj = act.retrievals.sonde.calculate_pbl_liu_liang(obj)

    obj = act.io.armfiles.read_netcdf(files[0])
    obj['tdry'].attrs['units'] = 'degree_Celsius'
    temp = obj['tdry'].values
    temp[20:50] = 100.
    obj['tdry'].values = temp
    with np.testing.assert_raises(ValueError):
        obj = act.retrievals.sonde.calculate_pbl_liu_liang(obj)



"""
This module contains I/O operations for loading MPL files.

"""

import os
import shutil
import subprocess
import tempfile

import xarray as xr
import dask

from act.io.armfiles import check_arm_standards


if shutil.which('mpl2nc') is not None:
    MPLIMPORT = True
else:
    MPLIMPORT = False


def read_sigma_mplv5(filename, save_nc=False, out_nc_path=None, afterpulse=None,
                     dead_time=None, overlap=None, **kwargs):
    """
    Returns `xarray.Dataset` with stored data and metadata from a user-defined
    SIGMA MPL V5 files. File is converted to netCDF using mpl2nc an optional
    dependency.

    Parameters
    ----------
    filename : str
        Name of file(s) to read.  If multiple, ensure that they are sorted,
        otherwise they will not concat properly
    save_nc : bool
        Whether or not to save intermediate step nc file.
    out_nc_path : str
        Path to save intermediate step nc file.
    afterpulse : str
        File with afterpulse correction (.bin)
    dead_time : str
        File with dead time correction (.bin)
    overlap : str
        File with overlap correction (.bin)
    **kwargs : keywords
        Keywords to pass through to xarray.open_dataset().

    """
    if not MPLIMPORT:
        raise ImportError(
            'The module mpl2nc is not installed and is needed to read '
            'mpl binary files!')

    if isinstance(filename, str):
        filename = [filename]

    task = []
    for f in filename:
        task.append(dask.delayed(proc_sigma_mplv5_read)(f, save_nc=save_nc,
                    out_nc_path=out_nc_path, afterpulse=afterpulse, dead_time=dead_time,
                    overlap=overlap, **kwargs))

    results = dask.compute(*task)

    obj = xr.concat(results, 'time')

    return obj


def proc_sigma_mplv5_read(f, save_nc=False, out_nc_path=None, afterpulse=None,
                          dead_time=None, overlap=None, **kwargs):
    """
    Returns `xarray.Dataset` with stored data and metadata from a user-defined
    SIGMA MPL V5 files. File is converted to netCDF using mpl2nc an optional
    dependency.  Dask sub-routine for processing

    Parameters
    ----------
    filename : str
        Name of file to read.
    save_nc : bool
        Whether or not to save intermediate step nc file.
    out_nc_path : str
        Path to save intermediate step nc file.
    afterpulse : str
        File with afterpulse correction (.bin)
    dead_time : str
        File with dead time correction (.bin)
    overlap : str
        File with overlap correction (.bin)
    **kwargs : keywords
        Keywords to pass through to xarray.open_dataset().

    """

    datastream_name = '.'.join(
        f.split('/')[-1].split('.')[0:2])
    if '.bin' not in f:
        mpl = True
        tmpfile1 = tempfile.mkstemp(suffix='.bin', dir='.')[1]
        shutil.copyfile(f, tmpfile1)
        f = tmpfile1
    else:
        mpl = False

    # Set up call for mpl2nc to add in correction files
    call = 'mpl2nc'
    if afterpulse is not None:
        call += ' -a ' + afterpulse
    if dead_time is not None:
        call += ' -d ' + dead_time
    if overlap is not None:
        call += ' -o ' + overlap

    call += ' ' + f

    # Specify the output, will use a temporary file if no output specified
    if save_nc:
        if out_nc_path is None:
            raise ValueError(
                'You are using save_nc, please specify '
                'an out_nc_path')

        subprocess.call(call + ' ' + out_nc_path, shell=True)
        ds = xr.open_dataset(out_nc_path, **kwargs)
    else:
        tmpfile2 = tempfile.mkstemp(suffix='.nc', dir='.')[1]
        subprocess.call(call + ' ' + tmpfile2, shell=True)
        ds = xr.open_dataset(tmpfile2, **kwargs)
        os.remove(tmpfile2)

    # Calculate range in meters
    ds['range'] = 0.5 * ds.bin_time[0] * ds.c * (ds.range + 0.5)

    # Swap the coordinates to be time and range
    ds = ds.swap_dims({'profile': 'time'})
    ds = ds.assign_coords({'time': ds.time,
                           'range': ds.range})

    # Add metadata
    is_arm_file_flag = check_arm_standards(ds)
    if is_arm_file_flag == 0:
        ds.attrs['_datastream'] = datastream_name
    ds.attrs['_arm_standards_flag'] = is_arm_file_flag

    if mpl:
        os.remove(tmpfile1)

    return ds



"""
This module contains I/O operations for loading csv files.

"""

import pandas as pd
import pathlib

from .armfiles import check_arm_standards


def read_csv(filename, sep=',', engine='python', column_names=None,
             skipfooter=0, **kwargs):

    """
    Returns an `xarray.Dataset` with stored data and metadata from user-defined
    query of CSV files.

    Parameters
    ----------
    filenames : str or list
        Name of file(s) to read.
    sep : str
        The separator between columns in the csv file.
    column_names : list or None
        The list of column names in the csv file.
    verbose : bool
        If true, will print if a file is not found.

    Additional keyword arguments will be passed into pandas.read_csv.

    Returns
    -------
    act_obj : Object
        ACT dataset. Will be None if the file is not found.

    Examples
    --------
    This example will load the example sounding data used for unit testing:

    .. code-block:: python

        import act
        the_ds, the_flag = act.io.csvfiles.read(
            act.tests.sample_files.EXAMPLE_CSV_WILDCARD)

    """

    # Convert to string if filename is a pathlib or not a list
    if isinstance(filename, (pathlib.PurePath, str)):
        filename = [str(filename)]

    if isinstance(filename, list) and isinstance(filename[0], pathlib.PurePath):
        print('filename')
        filename = [str(ii) for ii in filename]

    # Read data using pandas read_csv one file at a time and append to
    # list. Then concatinate the list into one pandas dataframe.
    li = []
    for fl in filename:
        df = pd.read_csv(fl, sep=sep, names=column_names,
                         skipfooter=skipfooter, engine=engine, **kwargs)
        li.append(df)

    if len(li) == 1:
        df = li[0]
    else:
        df = pd.concat(li, axis=0, ignore_index=True)

    # Set Coordinates if there's a variable date_time
    if 'date_time' in df:
        df.date_time = df.date_time.astype('datetime64')
        df.time = df.date_time
        df = df.set_index('time')

    # Convert to xarray DataSet
    ds = df.to_xarray()

    # Set additional variables
    # Since we cannot assume a standard naming convention setting
    # file_date and file_time to the first time in the file
    x_coord = ds.coords.to_index().values[0]
    if isinstance(x_coord, str):
        x_coord_dt = pd.to_datetime(x_coord)
        ds.attrs['_file_dates'] = x_coord_dt.strftime('%Y%m%d')
        ds.attrs['_file_times'] = x_coord_dt.strftime('%H%M%S')

    # Check for standard ARM datastream name, if none, assume the file is ARM
    # standard format.
    is_arm_file_flag = check_arm_standards(ds)
    if is_arm_file_flag == 0:

        ds.attrs['_datastream'] = '.'.join(filename[0].split('/')[-1].split('.')[0:2])

    # Add additional attributes, site, standards flag, etc...
    ds.attrs['_site'] = str(ds.attrs['_datastream'])[0:3]
    ds.attrs['_arm_standards_flag'] = is_arm_file_flag

    return ds



from os import path
from setuptools import setup, find_packages
import sys
import versioneer


# NOTE: This file must remain Python 2 compatible for the foreseeable future,
# to ensure that we error out properly for people with outdated setuptools
# and/or pip.
min_version = (3, 6)
if sys.version_info < min_version:
    error = """
act does not support Python {0}.{1}.
Python {2}.{3} and above is required. Check your Python version like so:

python3 --version

This may be due to an out-of-date pip. Make sure you have pip >= 9.0.1.
Upgrade pip like so:

pip install --upgrade pip
""".format(*sys.version_info[:2], *min_version)
    sys.exit(error)

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.rst'), encoding='utf-8') as readme_file:
    readme = readme_file.read()

with open(path.join(here, 'requirements.txt')) as requirements_file:
    # Parse requirements.txt, ignoring any commented-out lines.
    requirements = [line for line in requirements_file.read().splitlines()
                    if not line.startswith('#')]


setup(
    name='act-atmos',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="Package for working with atmospheric time series datasets",
    long_description=readme,
    long_description_content_type='text/x-rst',
    author="Adam Theisen",
    author_email='atheisen@anl.gov',
    url='https://github.com/ARM-DOE/ACT',
    packages=find_packages(exclude=['docs']),
    entry_points={'console_scripts': []},
    include_package_data=True,
    package_data={'act': []},
    install_requires=requirements,
    license="BSD (3-clause)",
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
    ],
)
